from __future__ import annotations

import lonboard.traits as t
from lonboard.controls._base import BaseControl


class NavigationControl(BaseControl):
    """A deck.gl NavigationControl.

    Passing this to [`Map.controls`][lonboard.Map.controls] will add zoom and compass
    buttons to the map.
    """

    _control_type = t.Unicode("navigation")

    show_compass = t.Bool(allow_none=True, default_value=None)
    """Whether to show the compass button.

    Default `True`.
    """

    show_zoom = t.Bool(allow_none=True, default_value=None)
    """Whether to show the zoom buttons.

    Default `True`.
    """

    visualize_pitch = t.Bool(allow_none=True, default_value=None)
    """Whether to enable pitch visualization.

    This only has effect for Maplibre-driven maps (i.e. where
    [`MaplibreBasemap.mode`][lonboard.basemap.MaplibreBasemap.mode] is "overlaid" or
    "interleaved").

    Default `True`.
    """

    visualize_roll = t.Bool(allow_none=True, default_value=None)
    """Whether to enable roll visualization.

    This only has effect for Maplibre-driven maps (i.e. where
    [`MaplibreBasemap.mode`][lonboard.basemap.MaplibreBasemap.mode] is "overlaid" or
    "interleaved").

    Default `False`.
    """



from __future__ import annotations

from dataclasses import dataclass


class BaseViewState:
    """Base class for view states."""


@dataclass(frozen=True)
class MapViewState(BaseViewState):
    """State of a [MapView][lonboard.experimental.view.MapView]."""

    longitude: float = 0
    """longitude at the map center"""

    latitude: float = 10
    """latitude at the map center."""

    zoom: float = 0.5
    """zoom level."""

    pitch: float = 0
    """pitch angle in degrees. `0` is top-down."""

    bearing: float = 0
    """bearing angle in degrees. `0` is north."""

    max_zoom: float = 20
    """max zoom level."""

    min_zoom: float = 0
    """min zoom level."""

    max_pitch: float = 60
    """max pitch angle."""

    min_pitch: float = 0
    """min pitch angle."""


@dataclass(frozen=True)
class GlobeViewState(BaseViewState):
    """State of a [GlobeView][lonboard.experimental.view.GlobeView]."""

    longitude: float
    """longitude at the viewport center."""

    latitude: float
    """latitude at the viewport center."""

    zoom: float
    """zoom level."""

    max_zoom: float = 20
    """max zoom level. Default 20."""

    min_zoom: float = 0
    """min zoom level. Default 0."""


@dataclass(frozen=True)
class FirstPersonViewState(BaseViewState):
    """State of a [FirstPersonView][lonboard.experimental.view.FirstPersonView]."""

    longitude: float
    """longitude of the camera position."""

    latitude: float
    """latitude of the camera position."""

    position: tuple[float, float, float] = (0.0, 0.0, 0.0)
    """Meter offsets of the camera from the lng-lat anchor point."""

    bearing: float = 0.0
    """bearing angle in degrees. `0` is north."""

    pitch: float = 0.0
    """pitch angle in degrees. `0` is horizontal."""

    max_pitch: float = 90.0
    """max pitch angle. Default 90 (down)."""

    min_pitch: float = -90.0
    """min pitch angle. Default -90 (up)."""


@dataclass(frozen=True)
class OrthographicViewState(BaseViewState):
    """State of an [OrthographicView][lonboard.experimental.view.OrthographicView]."""

    target: tuple[float, float, float] = (0.0, 0.0, 0.0)
    """The world position at the center of the viewport."""

    zoom: float | tuple[float, float] = 0.0
    """The zoom level of the viewport.

    - `zoom: 0` maps one unit distance to one pixel on screen, and increasing `zoom` by `1` scales the same object to twice as large. For example `zoom: 1` is 2x the original size, `zoom: 2` is 4x, `zoom: 3` is 8x etc.

    To apply independent zoom levels to the X and Y axes, supply a tuple [zoomX, zoomY].

    Default 0.
    """

    min_zoom: float | None = None
    """The min zoom level of the viewport. Default -Infinity."""

    max_zoom: float | None = None
    """The max zoom level of the viewport. Default Infinity."""


@dataclass(frozen=True)
class OrbitViewState(BaseViewState):
    """State of an [OrbitView][lonboard.experimental.view.OrbitView]."""

    target: tuple[float, float, float] = (0.0, 0.0, 0.0)
    """The world position at the center of the viewport."""

    rotation_orbit: float = 0.0
    """Rotating angle around orbit axis. Default 0."""

    rotation_x: float = 0.0
    """Rotating angle around X axis. Default 0."""

    zoom: float = 0.0
    """The zoom level of the viewport.

    `zoom: 0` maps one unit distance to one pixel on screen, and increasing `zoom` by
    `1` scales the same object to twice as large.

    Default 0.
    """

    min_zoom: float | None = None
    """The min zoom level of the viewport. Default -Infinity."""

    max_zoom: float | None = None
    """The max zoom level of the viewport. Default Infinity."""

    min_rotation_x: float = -90.0
    """The min rotating angle around X axis. Default -90."""

    max_rotation_x: float = 90.0
    """The max rotating angle around X axis. Default 90."""



# ruff: noqa: ARG002, SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from arro3.core import Array, ChunkedArray, DataType

from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    import pandas as pd

    from lonboard.layer import BaseArrowLayer


class FloatAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl float accessor.

    Various input is allowed:

    - An `int` or `float`. This will be used as the value for all objects.
    - A numpy `ndarray` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`FloatArray`][pyarrow.FloatArray], [`DoubleArray`][pyarrow.DoubleArray]
      or [`ChunkedArray`][pyarrow.ChunkedArray] containing either a `FloatArray` or
      `DoubleArray`. Each value in the array will be used as the value for the object at
      the same row index.
    - Any Arrow floating point array from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
      This includes data structures from
      [`geoarrow-rust`](https://geoarrow.github.io/geoarrow-rs/python/latest/).
    """

    default_value = float(0)
    info_text = (
        "a float value or numpy ndarray or Arrow array representing an array of floats"
    )

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _pandas_to_numpy(self, obj: BaseArrowLayer, value: pd.Series) -> np.ndarray:
        """Cast pandas Series to numpy ndarray."""
        return np.asarray(value)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.number):
            self.error(obj, value, info="numeric dtype")

        # TODO: should we always be casting to float32? Should it be
        # possible/allowed to pass in ~int8 or a data type smaller than float32?
        return ChunkedArray([value.astype(np.float32)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> float | ChunkedArray:
        if isinstance(value, (int, float, np.number)):
            return float(value)

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self._pandas_to_numpy(obj, value)

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_numeric(value.type):
            self.error(
                obj,
                value,
                info="Float Arrow array must be a numeric type.",
            )

        value = value.cast(DataType.float32())
        return value.rechunk(max_chunksize=obj._rows_per_chunk)



# ruff: noqa: D205

from __future__ import annotations

from typing import TYPE_CHECKING

import lonboard.traits as t
from lonboard._constants import EXTENSION_NAME
from lonboard.layer._base import BaseArrowLayer

if TYPE_CHECKING:
    import sys
    from typing import Self

    import duckdb
    import geopandas as gpd
    import pyproj
    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import ColumnLayerKwargs

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class ColumnLayer(BaseArrowLayer):
    """The ColumnLayer renders extruded cylinders (tessellated regular polygons) at given
    coordinates.
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[ColumnLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("column")

    table = t.ArrowTableTrait(allowed_geometry_types={EXTENSION_NAME.POINT})
    """A GeoArrow table with a Point or MultiPoint column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.ScatterplotLayer.from_geopandas] instead.
    """

    disk_resolution = t.Int(None, allow_none=True)
    """
    The number of sides to render the disk as. The disk is a regular polygon that fits
    inside the given radius. A higher resolution will yield a smoother look close-up,
    but also need more resources to render.

    - Type: `int`, optional
    - Default: `20`
    """

    radius = t.Float(None, allow_none=True)
    """
    Disk size in units specified by `radius_units` (default meters).

    - Type: `float`, optional
    - Default: `1000`
    """

    angle = t.Float(None, allow_none=True)
    """
    Disk rotation, counter-clockwise in degrees.

    - Type: `float`, optional
    - Default: `0`
    """

    offset = t.Tuple(t.Float(), t.Float(), default_value=None, allow_none=True).tag(
        sync=True,
    )
    """
    Disk offset from the position, relative to the radius. By default, the disk is
    centered at each position.

    - Type: `tuple[float, float]`, optional
    - Default: `(0, 0)`
    """

    coverage = t.Float(None, allow_none=True)
    """
    Radius multiplier, between 0 - 1. The radius of the disk is calculated by
    `coverage * radius`

    - Type: `float`, optional
    - Default: `1`
    """

    elevation_scale = t.Float(None, allow_none=True)
    """
    Column elevation multiplier. The elevation of column is calculated by
    `elevation_scale * get_elevation(d)`. `elevation_scale` is a handy property
    to scale all column elevations without updating the data.

    - Type: `float`, optional
    - Default: `1`
    """

    filled = t.Bool(None, allow_none=True)
    """
    Whether to draw a filled column (solid fill).

    - Type: `bool`, optional
    - Default: `True`
    """

    stroked = t.Bool(None, allow_none=True)
    """
    Whether to draw an outline around the disks. Only applies if `extruded=False`.

    - Type: `bool`, optional
    - Default: `False`
    """

    extruded = t.Bool(None, allow_none=True)
    """
    Whether to extrude the columns. If set to `false`, all columns will be rendered as
    flat polygons.

    - Type: `bool`, optional
    - Default: `True`
    """

    wireframe = t.Bool(None, allow_none=True)
    """
    Whether to generate a line wireframe of the column. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex around the disk. Only applies if `extruded=True`.

    - Type: `bool`, optional
    - Default: `False`
    """

    flat_shading = t.Bool(None, allow_none=True)
    """
    If `True`, the vertical surfaces of the columns use [flat
    shading](https://en.wikipedia.org/wiki/Shading#Flat_vs._smooth_shading). If `false`,
    use smooth shading. Only effective if `extruded` is `True`.

    - Type: `bool`, optional
    - Default: `False`
    """

    radius_units = t.Unicode(None, allow_none=True)
    """
    The units of the radius, one of `'meters'`, `'common'`, and `'pixels'`. See [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_units = t.Unicode(None, allow_none=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_scale = t.Float(None, allow_none=True, min=0)
    """
    The line width multiplier that multiplied to all outlines if the `stroked` attribute
    is `True`.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_min_pixels = t.Float(None, allow_none=True, min=0)
    """
    The minimum outline width in pixels. This can be used to prevent the line from
    getting too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    line_width_max_pixels = t.Float(None, allow_none=True, min=0)
    """
    The maximum outline width in pixels. This can be used to prevent the line from
    getting too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    get_fill_color = t.ColorAccessor(None, allow_none=True)
    """
    The filled color of each object in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the filled color for
          all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the filled color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = t.ColorAccessor(None, allow_none=True)
    """
    The outline color of each object in the format of `[r, g, b, [a]]`. Each channel is
    a number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the outline color for
          all objects.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the outline color for the object at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_elevation = t.FloatAccessor(None, allow_none=True)
    """
    The elevation of each cell in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """

    get_line_width = t.FloatAccessor(None, allow_none=True)
    """
    The width of the outline of each column, in units specified by `line_width_units`
    (default `'meters'`). Only applies if `extruded: false` and `stroked: true`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the outline width for all columns.
        - If an array is provided, each value in the array will be used as the outline
          width for the column at the same row index.
    - Default: `1`.
    """



"""New traits for our custom serialization.

Refer to https://traitlets.readthedocs.io/en/stable/defining_traits.html for
documentation on how to define new traitlet types.
"""

from ._a5 import A5Accessor
from ._base import FixedErrorTraitType, VariableLengthTuple
from ._color import ColorAccessor
from ._extensions import DashArrayAccessor, FilterCategoryAccessor, FilterValueAccessor
from ._float import FloatAccessor
from ._h3 import H3Accessor
from ._map import BasemapUrl, MapHeightTrait, ViewStateTrait
from ._normal import NormalAccessor
from ._point import PointAccessor
from ._table import ArrowTableTrait
from ._text import TextAccessor
from ._timestamp import TimestampAccessor
from ._upstream_wrappers import Any, Bool, Enum, Float, Int, List, Tuple, Unicode, Union

__all__ = [
    "A5Accessor",
    "Any",
    "ArrowTableTrait",
    "BasemapUrl",
    "Bool",
    "ColorAccessor",
    "DashArrayAccessor",
    "Enum",
    "FilterCategoryAccessor",
    "FilterValueAccessor",
    "FixedErrorTraitType",
    "Float",
    "FloatAccessor",
    "H3Accessor",
    "Int",
    "List",
    "MapHeightTrait",
    "NormalAccessor",
    "PointAccessor",
    "TextAccessor",
    "TimestampAccessor",
    "Tuple",
    "Unicode",
    "Union",
    "VariableLengthTuple",
    "ViewStateTrait",
]



import lonboard.traits as t
from lonboard._base import BaseWidget
from lonboard.view_state import (
    BaseViewState,
    FirstPersonViewState,
    GlobeViewState,
    MapViewState,
    OrbitViewState,
    OrthographicViewState,
)


class BaseView(BaseWidget):
    """A deck.gl [View][deck-view].

    [deck-view]: https://deck.gl/docs/api-reference/core/view

    The `BaseView` class and its subclasses are used to specify where and how your deck.gl layers should be rendered. Applications typically instantiate at least one `BaseView` subclass.
    """

    _view_state_type: type[BaseViewState] = BaseViewState

    x = t.Union([t.Int(), t.Unicode()], allow_none=True, default_value=None).tag(
        sync=True,
    )
    """The x position of the view.

    A relative (e.g. `'50%'`) or absolute position. Default `0`.
    """

    y = t.Union([t.Int(), t.Unicode()], allow_none=True, default_value=None).tag(
        sync=True,
    )
    """The y position of the view.

    A relative (e.g. `'50%'`) or absolute position. Default `0`.
    """

    width = t.Union([t.Int(), t.Unicode()], allow_none=True, default_value=None).tag(
        sync=True,
    )
    """The width of the view.

    A relative (e.g. `'50%'`) or absolute extent. Default `'100%'`.
    """

    height = t.Union([t.Int(), t.Unicode()], allow_none=True, default_value=None).tag(
        sync=True,
    )
    """The height of the view.

    A relative (e.g. `'50%'`) or absolute extent. Default `'100%'`.
    """


class MapView(BaseView):
    """A deck.gl [MapView][deck-map-view].

    [deck-map-view]: https://deck.gl/docs/api-reference/core/map-view

    The `MapView` creates a camera that looks at a geospatial location on a map from a certain direction. The behavior of `MapView` is generally modeled after that of Mapbox GL JS.

    Most geospatial applications will use this view with the default parameters.
    """

    _view_type = t.Unicode("map-view")

    _view_state_type = MapViewState

    repeat = t.Bool(allow_none=True, default_value=None)
    """
    Whether to render multiple copies of the map at low zoom levels. Default `false`.
    """

    near_z_multiplier = t.Float(allow_none=True, default_value=None)
    """Scaler for the near plane, 1 unit equals to the height of the viewport.

    Default to `0.1`. Overwrites the `near` parameter.
    """

    far_z_multiplier = t.Float(allow_none=True, default_value=None)
    """Scaler for the far plane, 1 unit equals to the distance from the camera to the top edge of the screen.

    Default to `1.01`. Overwrites the `far` parameter.
    """

    projection_matrix = t.List(
        t.Float(),
        allow_none=True,
        default_value=None,
        minlen=16,
        maxlen=16,
    ).tag(
        sync=True,
    )
    """Projection matrix.

    If `projectionMatrix` is not supplied, the `View` class will build a projection matrix from the following parameters:
    """

    fovy = t.Float(allow_none=True, default_value=None)
    """Field of view covered by camera, in the perspective case. In degrees.

    If not supplied, will be calculated from `altitude`.
    """

    altitude = t.Float(allow_none=True, default_value=None)
    """Distance of the camera relative to viewport height.

    Default `1.5`.
    """

    orthographic = t.Bool(allow_none=True, default_value=None)
    """Whether to create an orthographic or perspective projection matrix.

    Default is `false` (perspective projection).
    """


class GlobeView(BaseView):
    """A deck.gl GlobeView.

    [deck-globe-view]: https://deck.gl/docs/api-reference/core/globe-view

    The `GlobeView` projects the earth into a 3D globe.
    """

    _view_type = t.Unicode("globe-view")

    _view_state_type = GlobeViewState

    resolution = t.Float(allow_none=True, default_value=None)
    """The resolution at which to turn flat features into 3D meshes, in degrees.

    Smaller numbers will generate more detailed mesh. Default `10`.
    """

    near_z_multiplier = t.Float(allow_none=True, default_value=None)
    """Scaler for the near plane, 1 unit equals to the height of the viewport.

    Default to `0.1`. Overwrites the `near` parameter.
    """

    far_z_multiplier = t.Float(allow_none=True, default_value=None)
    """Scaler for the far plane, 1 unit equals to the distance from the camera to the top edge of the screen.

    Default to `2`. Overwrites the `far` parameter.
    """


class FirstPersonView(BaseView):
    """A deck.gl FirstPersonView.

    The `FirstPersonView` class is a subclass of `View` that describes a camera placed at a provided location, looking towards the direction and orientation specified by viewState. The behavior is similar to that of a first-person game.
    """

    _view_type = t.Unicode("first-person-view")

    _view_state_type = FirstPersonViewState

    projection_matrix = t.List(
        t.Float(),
        allow_none=True,
        default_value=None,
        minlen=16,
        maxlen=16,
    ).tag(
        sync=True,
    )
    """Projection matrix.

    If `projectionMatrix` is not supplied, the `View` class will build a projection matrix from the following parameters:
    """

    fovy = t.Float(allow_none=True, default_value=None)
    """Field of view covered by camera, in the perspective case. In degrees.

    Default `50`.
    """

    near = t.Float(allow_none=True, default_value=None)
    """Distance of near clipping plane.

    Default `0.1`.
    """

    far = t.Float(allow_none=True, default_value=None)
    """Distance of far clipping plane.

    Default `1000`.
    """

    focal_distance = t.Float(allow_none=True, default_value=None)
    """Modifier of viewport scale.

    Corresponds to the number of pixels per meter. Default `1`.
    """


class OrbitView(BaseView):
    """A deck.gl OrbitView.

    The `OrbitView` class is a subclass of `View` that creates a 3D camera that rotates around a target position. It is usually used for the examination of a 3D scene in non-geospatial use cases.
    """

    _view_type = t.Unicode("orbit-view")

    _view_state_type = OrbitViewState

    orbit_axis = t.Unicode(allow_none=True, default_value=None)
    """Axis with 360 degrees rotating freedom, either `'Y'` or `'Z'`, default to `'Z'`."""

    projection_matrix = t.List(
        t.Float(),
        allow_none=True,
        default_value=None,
        minlen=16,
        maxlen=16,
    ).tag(
        sync=True,
    )
    """Projection matrix.

    If `projectionMatrix` is not supplied, the `View` class will build a projection matrix from the following parameters:
    """

    fovy = t.Float(allow_none=True, default_value=None)
    """Field of view covered by camera, in the perspective case. In degrees.

    Default `50`.
    """

    near = t.Float(allow_none=True, default_value=None)
    """Distance of near clipping plane.

    Default `0.1`.
    """

    far = t.Float(allow_none=True, default_value=None)
    """Distance of far clipping plane.

    Default `1000`.
    """

    orthographic = t.Bool(allow_none=True, default_value=None)
    """Whether to create an orthographic or perspective projection matrix.

    Default is `false` (perspective projection).
    """


class OrthographicView(BaseView):
    """A deck.gl OrthographicView.

    The `OrthographicView` class is a subclass of `View` that creates a top-down view of the XY plane. It is usually used for rendering 2D charts in non-geospatial use cases.
    """

    _view_type = t.Unicode("orthographic-view")

    _view_state_type = OrthographicViewState

    flip_y = t.Bool(allow_none=True, default_value=None)
    """
    Whether to use top-left coordinates (`true`) or bottom-left coordinates (`false`).

    Default `true`.
    """

    near = t.Float(allow_none=True, default_value=None)
    """Distance of near clipping plane.

    Default `0.1`.
    """

    far = t.Float(allow_none=True, default_value=None)
    """Distance of far clipping plane.

    Default `1000`.
    """



from __future__ import annotations

import lonboard.traits as t
from lonboard._base import BaseWidget


class BaseControl(BaseWidget):
    """A deck.gl or Maplibre Control."""

    position = t.Union(
        [
            t.Unicode("top-left"),
            t.Unicode("top-right"),
            t.Unicode("bottom-left"),
            t.Unicode("bottom-right"),
        ],
        allow_none=True,
        default_value=None,
    )
    """Position of the control in the map.

    One of `"top-left"`, `"top-right"`, `"bottom-left"`, or `"bottom-right"`.
    """



"""Available layer types that can be rendered on a Lonboard Map."""

# Notes:
# - When we pass a value of `None` as a default value to a trait, that value will be
#   serialized to JS as `null` and will not be passed into the GeoArrow model (see the
#   lengthy assignments of type `..(isDefined(this.param) && { param: this.param })`).
#   Then the default value in the JS GeoArrow layer (defined in
#   `@geoarrow/deck.gl-layers`) will be used.

from ._a5 import A5Layer
from ._arc import ArcLayer
from ._base import BaseArrowLayer, BaseLayer
from ._bitmap import BitmapLayer, BitmapTileLayer
from ._column import ColumnLayer
from ._geohash import GeohashLayer
from ._h3 import H3HexagonLayer
from ._heatmap import HeatmapLayer
from ._path import PathLayer
from ._point_cloud import PointCloudLayer
from ._polygon import PolygonLayer, SolidPolygonLayer
from ._raster import RasterLayer
from ._s2 import S2Layer
from ._scatterplot import ScatterplotLayer
from ._trips import TripsLayer

__all__ = [
    "A5Layer",
    "ArcLayer",
    "BaseArrowLayer",
    "BaseLayer",
    "BitmapLayer",
    "BitmapTileLayer",
    "ColumnLayer",
    "GeohashLayer",
    "H3HexagonLayer",
    "HeatmapLayer",
    "PathLayer",
    "PointCloudLayer",
    "PolygonLayer",
    "RasterLayer",
    "S2Layer",
    "ScatterplotLayer",
    "SolidPolygonLayer",
    "TripsLayer",
]



"""Implement h3 cell validation in pure numpy.

It's hard to surface errors from deck.gl back to Python, so it's a bad user experience
if the JS console errors and silently nothing renders. But also I don't want to depend
on the h3 library for this because the h3 library isn't vectorized (arghhhh!) and I
don't want to require the dependency.

So instead, I spend my time porting code into Numpy 😄.

Ported from Rust code in h3o:

https://github.com/HydroniumLabs/h3o/blob/07dcb85d9cb539f685ec63050ef0954b1d9f3864/src/index/cell.rs#L1897-L1962
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = ["validate_h3_indices"]

MODE_OFFSET = 59
"""Offset (in bits) of the mode in an H3 index."""

MODE_MASK = 0b1111 << MODE_OFFSET

EDGE_OFFSET = 56
"""Offset (in bits) of the cell edge in an H3 index."""

EDGE_MASK = 0b111 << EDGE_OFFSET

VERTEX_OFFSET = 56
"""Offset (in bits) of the cell vertex in an H3 index."""

VERTEX_MASK = 0b111 << VERTEX_OFFSET

DIRECTIONS_MASK = 0x0000_1FFF_FFFF_FFFF
"""Bitmask to select the directions bits in an H3 index."""

INDEX_MODE_CELL = 1
"""H3 index mode for cells."""

BASE_CELL_OFFSET = 45
"""Offset (in bits) of the base cell in an H3 index."""

BASE_CELL_MASK = 0b111_1111 << BASE_CELL_OFFSET
"""Bitmask to select the base cell bits in an H3 index."""

MAX_BASE_CELL = 121
"""Maximum value for a base cell."""

RESOLUTION_OFFSET = 52
"""The bit offset of the resolution in an H3 index."""

RESOLUTION_MASK = 0b1111 << RESOLUTION_OFFSET
"""Bitmask to select the resolution bits in an H3 index."""

MAX_RESOLUTION = 15
"""Maximum supported H3 resolution."""

DIRECTION_BITSIZE = 3
"""Size, in bits, of a direction (range [0; 6])."""

BASE_PENTAGONS_HI = 0x0020_0802_0008_0100
"""Bitmap where a bit's position represents a base cell value (high part).

Refactored from upstream 128 bit integer
https://github.com/HydroniumLabs/h3o/blob/3b40550291a57552117c48c19841557a3b0431e1/src/base_cell.rs#L12
"""

BASE_PENTAGONS_LO = 0x8402_0040_0100_4010
"""Bitmap where a bit's position represents a base cell value (low part).

Refactored from upstream 128 bit integer
https://github.com/HydroniumLabs/h3o/blob/3b40550291a57552117c48c19841557a3b0431e1/src/base_cell.rs#L12
"""

PENTAGON_BASE_CELLS = np.array(
    [4, 14, 24, 33, 38, 49, 58, 63, 72, 83, 97, 107],
    dtype=np.uint8,
)
"""Set of pentagon base cells."""


def validate_h3_indices(h3_indices: NDArray[np.uint64]) -> None:
    """Validate an array of uint64 H3 indices.

    Raises ValueError if any index is invalid.
    """
    invalid_reserved_bits = h3_indices >> 56 & 0b1000_0111 != 0
    bad_indices = np.where(invalid_reserved_bits)[0]
    if len(bad_indices) > 0:
        raise ValueError(
            f"Tainted reserved bits in indices: {bad_indices.tolist()}\n"
            f"with values {h3_indices[bad_indices].tolist()}",
        )

    invalid_mode = get_mode(h3_indices) != INDEX_MODE_CELL
    bad_indices = np.where(invalid_mode)[0]
    if len(bad_indices) > 0:
        raise ValueError(
            f"Invalid index mode in indices: {bad_indices.tolist()}",
            f"with values {h3_indices[bad_indices].tolist()}",
        )

    base = get_base_cell(h3_indices)
    invalid_base_cell = base > MAX_BASE_CELL
    bad_indices = np.where(invalid_base_cell)[0]
    if len(bad_indices) > 0:
        raise ValueError(
            f"Invalid base cell in indices: {bad_indices.tolist()}",
            f"with values {h3_indices[bad_indices].tolist()}",
        )

    # Resolution is always valid: coded on 4 bits, valid range is [0; 15].
    resolution = get_resolution(h3_indices)

    # Check that we have a tail of unused cells  after `resolution` cells.
    #
    # We expect every bit to be 1 in the tail (because unused cells are
    # represented by `0b111`), i.e. every bit set to 0 after a NOT.
    unused_count = MAX_RESOLUTION - resolution
    unused_bitsize = unused_count * DIRECTION_BITSIZE
    unused_mask = (1 << unused_bitsize.astype(np.uint64)) - 1
    invalid_unused_direction_pattern = (~h3_indices) & unused_mask != 0
    bad_indices = np.where(invalid_unused_direction_pattern)[0]
    if len(bad_indices) > 0:
        raise ValueError(
            f"Invalid unused direction pattern in indices: {bad_indices.tolist()}",
            f"with values {h3_indices[bad_indices].tolist()}",
        )

    # Check that we have `resolution` valid cells (no unused ones).
    dirs_mask = (1 << (resolution * DIRECTION_BITSIZE).astype(np.uint64)) - 1
    dirs = (h3_indices >> unused_bitsize) & dirs_mask
    invalid_unused_direction = has_unused_direction(dirs)
    bad_indices = np.where(invalid_unused_direction)[0]
    if len(bad_indices) > 0:
        raise ValueError(
            f"Unexpected unused direction in indices: {bad_indices.tolist()}",
            f"with values {h3_indices[bad_indices].tolist()}",
        )

    # Check for pentagons with deleted subsequence.
    has_pentagon_base = np.logical_and(is_pentagon(base), resolution != 0)
    pentagon_base_indices = np.where(has_pentagon_base)[0]
    if len(pentagon_base_indices) > 0:
        pentagons = h3_indices[pentagon_base_indices]
        pentagon_resolutions = resolution[pentagon_base_indices]
        pentagon_dirs = dirs[pentagon_base_indices]

        # Move directions to the front, so that we can count leading zeroes.
        pentagon_offset = 64 - (pentagon_resolutions * DIRECTION_BITSIZE)

        # NOTE: The following was ported via GPT from Rust `leading_zeros`
        # https://github.com/HydroniumLabs/h3o/blob/07dcb85d9cb539f685ec63050ef0954b1d9f3864/src/index/cell.rs#L1951

        # Find the position of the first bit set, if it's a multiple of 3
        # that means we have a K axe as the first non-center direction,
        # which is forbidden.
        shifted = pentagon_dirs << pentagon_offset

        # Compute leading zeros for each element (assuming 64-bit unsigned integers)
        # where `leading_zeros = 64 - shifted.bit_length()`
        # numpy doesn't have bit_length, so use log2 and handle zeros
        bitlen = np.where(shifted == 0, 0, np.floor(np.log2(shifted)).astype(int) + 1)
        leading_zeros = 64 - bitlen

        # Add 1 and check if multiple of 3
        is_multiple_of_3 = ((leading_zeros + 1) % 3) == 0
        bad_indices = np.where(is_multiple_of_3)[0]
        if len(bad_indices) > 0:
            raise ValueError(
                f"Pentagonal cell index with a deleted subsequence: {bad_indices.tolist()}",
                f"with values {pentagons[bad_indices].tolist()}",
            )


def get_mode(bits: NDArray[np.uint64]) -> NDArray[np.uint8]:
    """Return the H3 index mode bits."""
    return ((bits & MODE_MASK) >> MODE_OFFSET).astype(np.uint8)


def get_base_cell(bits: NDArray[np.uint64]) -> NDArray[np.uint8]:
    """Return the H3 index base cell bits."""
    return ((bits & BASE_CELL_MASK) >> BASE_CELL_OFFSET).astype(np.uint8)


def get_resolution(bits: NDArray[np.uint64]) -> NDArray[np.uint8]:
    """Return the H3 index resolution."""
    return ((bits & RESOLUTION_MASK) >> RESOLUTION_OFFSET).astype(np.uint8)


def has_unused_direction(dirs: NDArray) -> NDArray[np.bool_]:
    """Check if there is at least one unused direction in the given directions.

    Copied from upstream
    https://github.com/HydroniumLabs/h3o/blob/07dcb85d9cb539f685ec63050ef0954b1d9f3864/src/index/cell.rs#L2056-L2107
    """
    LO_MAGIC = 0b001_001_001_001_001_001_001_001_001_001_001_001_001_001_001  # noqa: N806
    HI_MAGIC = 0b100_100_100_100_100_100_100_100_100_100_100_100_100_100_100  # noqa: N806

    return ((~dirs - LO_MAGIC) & (dirs & HI_MAGIC)) != 0


def is_pentagon(cell: NDArray[np.uint8]) -> NDArray[np.bool_]:
    """Return true if the base cell is pentagonal.

    Note that this is **not** copied from the upstream:
    https://github.com/HydroniumLabs/h3o/blob/3b40550291a57552117c48c19841557a3b0431e1/src/base_cell.rs#L33-L47

    Because they use a 128 bit integer as a bitmap, which is not available in
    numpy. Instead we use a simple lookup in a static array.
    """
    return np.isin(cell, PENTAGON_BASE_CELLS)



# ruff: noqa: ARG002, SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from arro3.core import Array, ChunkedArray, DataType

from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    import pandas as pd

    from lonboard.layer import BaseArrowLayer


class TextAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl text accessor.

    Various input is allowed:

    - A `str`. This will be used as the value for all objects.
    - A numpy `ndarray` with a string data type. Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a string data type. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`StringArray`][pyarrow.StringArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing a `StringArray`. Each value in
      the array will be used as the value for the object at the same row index.
    """

    default_value = ""
    info_text = (
        "a string value or numpy ndarray or pandas Series or Arrow array representing"
        " an array of strings"
    )

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _pandas_to_arrow(self, obj: BaseArrowLayer, value: pd.Series) -> ChunkedArray:
        """Cast pandas Series to arrow array."""
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow is a required dependency when passing in a pandas series",
            ) from e

        return ChunkedArray([pa.array(value)])

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow is a required dependency when passing in a numpy string array",
            ) from e

        return ChunkedArray([pa.StringArray.from_pandas(value)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> float | str | ChunkedArray:
        if isinstance(value, str):
            return value

        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self._pandas_to_arrow(obj, value)
        elif isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if DataType.is_large_string(value.type):
            value = value.cast(DataType.string())

        if not DataType.is_string(value.type):
            self.error(
                obj,
                value,
                info="String Arrow array must be a string type.",
            )

        return value.rechunk(max_chunksize=obj._rows_per_chunk)



# ruff: noqa: D205, ERA001

from __future__ import annotations

from typing import TYPE_CHECKING, Unpack

import lonboard.traits as t
from lonboard._geoarrow.ops import Bbox, WeightedCentroid
from lonboard.layer._base import BaseLayer

if TYPE_CHECKING:
    from lonboard.types.layer import BitmapLayerKwargs, BitmapTileLayerKwargs


class BitmapLayer(BaseLayer):
    """The `BitmapLayer` renders a bitmap (e.g. PNG, JPEG, or WebP) at specified
    boundaries.

    **Example:**

    ```py
    from lonboard import Map, BitmapLayer

    layer = BitmapLayer(
        image='https://raw.githubusercontent.com/visgl/deck.gl-data/master/website/sf-districts.png',
        bounds=[-122.5190, 37.7045, -122.355, 37.829]
    )
    m = Map(layer)
    m
    ```
    """

    def __init__(self, **kwargs: Unpack[BitmapLayerKwargs]) -> None:
        super().__init__(**kwargs)  # type: ignore

    _layer_type = t.Unicode("bitmap")

    image = t.Unicode()
    """The URL to an image to display.

    - Type: `str`
    """

    bounds = t.Union(
        [
            t.VariableLengthTuple(t.Float(), minlen=4, maxlen=4),
            t.VariableLengthTuple(
                t.VariableLengthTuple(t.Float(), minlen=2, maxlen=2),
                minlen=4,
                maxlen=4,
            ),
        ],
    )
    """The bounds of the image.

    Supported formats:

        - Coordinates of the bounding box of the bitmap `[left, bottom, right, top]`
        - Coordinates of four corners of the bitmap, should follow the sequence of
          `[[left, bottom], [left, top], [right, top], [right, bottom]]`.
    """

    desaturate = t.Float(0, min=0, max=1)
    """The desaturation of the bitmap. Between `[0, 1]`.

    - Type: `float`, optional
    - Default: `0`
    """

    transparent_color = t.VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to use for transparent pixels, in `[r, g, b, a]`.

    - Type: `List[float]`, optional
    - Default: `[0, 0, 0, 0]`
    """

    tint_color = t.VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to tint the bitmap by, in `[r, g, b]`.

    - Type: `List[float]`, optional
    - Default: `[255, 255, 255]`
    """

    @property
    def _bbox(self) -> Bbox:
        a, b, c, d = self.bounds

        # Four corners
        if isinstance(a, tuple):
            bbox = Bbox()
            bbox.update(Bbox(a[0], a[1], a[0], a[1]))
            bbox.update(Bbox(b[0], b[1], b[0], b[1]))
            bbox.update(Bbox(c[0], c[1], c[0], c[1]))
            bbox.update(Bbox(d[0], d[1], d[0], d[1]))
            return bbox

        return Bbox(a, b, c, d)

    @property
    def _weighted_centroid(self) -> WeightedCentroid:
        bbox = self._bbox
        center_x = (bbox.minx + bbox.maxx) / 2
        center_y = (bbox.miny + bbox.maxy) / 2
        # no idea what weight to put on this; we don't know how many "objects" this
        # image should represent.
        return WeightedCentroid(x=center_x, y=center_y, num_items=100)


class BitmapTileLayer(BaseLayer):
    """The BitmapTileLayer renders image tiles (e.g. PNG, JPEG, or WebP) in the web
    mercator tiling system. Only the tiles visible in the current viewport are loaded
    and rendered.

    **Example:**

    ```py
    from lonboard import Map, BitmapTileLayer

    # We set `max_requests < 0` because `tile.openstreetmap.org` supports HTTP/2.
    layer = BitmapTileLayer(
        data="https://tile.openstreetmap.org/{z}/{x}/{y}.png",
        tile_size=256,
        max_requests=-1,
        min_zoom=0,
        max_zoom=19,
    )
    m = Map(layer)
    ```
    """

    def __init__(self, **kwargs: Unpack[BitmapTileLayerKwargs]) -> None:
        super().__init__(**kwargs)  # type: ignore

    _layer_type = t.Unicode("bitmap-tile")

    data = t.Union([t.Unicode(), t.VariableLengthTuple(t.Unicode(), minlen=1)]).tag(
        sync=True,
    )
    """
    Either a URL template or an array of URL templates from which the tile data should
    be loaded.

    If the value is a string: a URL template. Substrings {x} {y} and {z}, if present,
    will be replaced with a tile's actual index when it is requested.

    If the value is an array: multiple URL templates. Each endpoint must return the same
    content for the same tile index. This can be used to work around domain sharding,
    allowing browsers to download more resources simultaneously. Requests made are
    balanced among the endpoints, based on the tile index.
    """

    tile_size = t.Int(None, allow_none=True)
    """
    The pixel dimension of the tiles, usually a power of 2.

    Tile size represents the target pixel width and height of each tile when rendered.
    Smaller tile sizes display the content at higher resolution, while the layer needs
    to load more tiles to fill the same viewport.

    - Type: `int`, optional
    - Default: `512`
    """

    zoom_offset = t.Int(None, allow_none=True)
    """
    This offset changes the zoom level at which the tiles are fetched. Needs to be an
    integer.

    - Type: `int`, optional
    - Default: `0`
    """

    max_zoom = t.Int(None, allow_none=True)
    """
    The max zoom level of the layer's data. When overzoomed (i.e. `zoom > max_zoom`),
    tiles from this level will be displayed.

    - Type: `int`, optional
    - Default: `None`
    """

    min_zoom = t.Int(None, allow_none=True)
    """
    The min zoom level of the layer's data. When underzoomed (i.e. `zoom < min_zoom`),
    the layer will not display any tiles unless `extent` is defined, to avoid issuing
    too many tile requests.

    - Type: `int`, optional
    - Default: `None`
    """

    extent = t.VariableLengthTuple(
        t.Float(),
        minlen=4,
        maxlen=4,
        allow_none=True,
        default_value=None,
    )
    """
    The bounding box of the layer's data, in the form of `[min_x, min_y, max_x, max_y]`.
    If provided, the layer will only load and render the tiles that are needed to fill
    this box.

    - Type: `List[float]`, optional
    - Default: `None`
    """

    max_cache_size = t.Int(None, allow_none=True)
    """
    The maximum number of tiles that can be cached. The tile cache keeps loaded tiles in
    memory even if they are no longer visible. It reduces the need to re-download the
    same data over and over again when the user pan/zooms around the map, providing a
    smoother experience.

    If not supplied, the `max_cache_size` is calculated as 5 times the number of tiles
    in the current viewport.

    - Type: `int`, optional
    - Default: `None`
    """

    # TODO: Not sure if `getTileData` returns a `byteLength`?
    # max_cache_byte_size = t.Int(None, allow_none=True)
    # """
    # """

    refinement_strategy = t.Unicode(None, allow_none=True)
    """How the tile layer refines the visibility of tiles.

    When zooming in and out, if the layer only shows tiles from the current zoom level,
    then the user may observe undesirable flashing while new data is loading. By setting
    `refinement_strategy` the layer can attempt to maintain visual continuity by
    displaying cached data from a different zoom level before data is available.

    This prop accepts one of the following:

    - `"best-available"`: If a tile in the current viewport is waiting for its data to
      load, use cached content from the closest zoom level to fill the empty space. This
      approach minimizes the visual flashing due to missing content.
    - `"no-overlap"`: Avoid showing overlapping tiles when backfilling with cached
      content. This is usually favorable when tiles do not have opaque backgrounds.
    - `"never"`: Do not display any tile that is not selected.

    - Type: `str`, optional
    - Default: `"best-available"`
    """

    max_requests = t.Int(None, allow_none=True)
    """The maximum number of concurrent data fetches.

    If <= 0, no throttling will occur, and `get_tile_data` may be called an unlimited
    number of times concurrently regardless of how long that tile is or was visible.

    If > 0, a maximum of `max_requests` instances of `get_tile_data` will be called
    concurrently. Requests may never be called if the tile wasn't visible long enough to
    be scheduled and started. Requests may also be aborted (through the signal passed to
    `get_tile_data`) if there are more than `max_requests` ongoing requests and some of
    those are for tiles that are no longer visible.

    If `get_tile_data` makes fetch requests against an HTTP 1 web server, then
    max_requests should correlate to the browser's maximum number of concurrent fetch
    requests. For Chrome, the max is 6 per domain. If you use the data prop and specify
    multiple domains, you can increase this limit. For example, with Chrome and 3
    domains specified, you can set max_requests=18.

    If the web server supports HTTP/2 (Open Chrome dev tools and look for "h2" in the
    Protocol column), then you can make an unlimited number of concurrent requests (and
    can set max_requests=-1). Note that this will request data for every tile, no matter
    how long the tile was visible, and may increase server load.
    """

    desaturate = t.Float(0, min=0, max=1)
    """The desaturation of the bitmap. Between `[0, 1]`.

    - Type: `float`, optional
    - Default: `0`
    """

    transparent_color = t.VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to use for transparent pixels, in `[r, g, b, a]`.

    - Type: `List[float]`, optional
    - Default: `[0, 0, 0, 0]`
    """

    tint_color = t.VariableLengthTuple(
        t.Float(),
        default_value=None,
        allow_none=True,
        minlen=3,
        maxlen=4,
    )
    """The color to tint the bitmap by, in `[r, g, b]`.

    - Type: `List[float]`, optional
    - Default: `[255, 255, 255]`
    """



# ruff: noqa: SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from typing import cast as type_cast

from arro3.core import DataType, Table

from lonboard._constants import EXTENSION_NAME
from lonboard._geoarrow.box_to_polygon import parse_box_encoded_table
from lonboard._serialization import TABLE_SERIALIZATION
from lonboard._utils import get_geometry_column_index
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    from lonboard.layer import BaseArrowLayer


class ArrowTableTrait(FixedErrorTraitType):
    """A trait to validate input for a geospatial Arrow-backed table.

    Allowed input includes:

    - A pyarrow [`Table`][pyarrow.Table] or containing a geometry column with [GeoArrow metadata](https://geoarrow.org/extension-types).
    - Any GeoArrow table from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = None
    info_text = (
        "a table-like Arrow object, such as a pyarrow or arro3 Table or "
        "RecordBatchReader"
    )

    def __init__(
        self,
        *args: Any,
        allowed_geometry_types: set[EXTENSION_NAME] | None = None,
        allowed_dimensions: set[int] | None = None,
        geometry_required: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(
            sync=True,
            allowed_geometry_types=allowed_geometry_types,
            allowed_dimensions=allowed_dimensions,
            geometry_required=geometry_required,
            **TABLE_SERIALIZATION,
        )

    def validate(self, obj: BaseArrowLayer, value: Any) -> Table:
        if not isinstance(value, Table):
            self.error(obj, value)

        allowed_geometry_types = self.metadata.get("allowed_geometry_types")
        allowed_geometry_types = type_cast(
            "set[bytes] | None",
            allowed_geometry_types,
        )

        allowed_dimensions = self.metadata.get("allowed_dimensions")
        allowed_dimensions = type_cast("set[int] | None", allowed_dimensions)

        geom_col_idx = get_geometry_column_index(value.schema)

        geometry_required = self.metadata.get("geometry_required")
        if geometry_required and geom_col_idx is None:
            return self.error(obj, value, info="geometry column in table")

        # No restriction on the allowed geometry types in this table
        if allowed_geometry_types:
            # If we allow polygons as input, then we also allow geoarrow.box.
            # Convert boxes to Polygons
            if EXTENSION_NAME.POLYGON in allowed_geometry_types:
                value = parse_box_encoded_table(value)

            geometry_extension_type = value.schema.field(geom_col_idx).metadata.get(
                b"ARROW:extension:name",
            )

            if (
                allowed_geometry_types
                and geometry_extension_type not in allowed_geometry_types
            ):
                allowed_types_str = ", ".join(map(str, allowed_geometry_types))
                msg = (
                    f"Expected one of {allowed_types_str} geometry types, "
                    f"got {geometry_extension_type}."
                )
                self.error(obj, value, info=msg)

        if allowed_dimensions:
            typ = value.column(geom_col_idx).type
            while DataType.is_list(typ):
                value_type = typ.value_type
                assert value_type is not None
                typ = value_type

            assert DataType.is_fixed_size_list(typ)
            if typ.list_size not in allowed_dimensions:
                msg = " or ".join(map(str, list(allowed_dimensions)))
                self.error(obj, value, info=f"{msg}-dimensional points")

        return value.rechunk(max_chunksize=obj._rows_per_chunk)



from __future__ import annotations

from typing import TYPE_CHECKING

import lonboard.traits as t
from lonboard._constants import EXTENSION_NAME
from lonboard.layer._base import BaseArrowLayer

if TYPE_CHECKING:
    import sys
    from typing import Self

    import duckdb
    import geopandas as gpd
    import pyproj
    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import PolygonLayerKwargs, SolidPolygonLayerKwargs

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class PolygonLayer(BaseArrowLayer):
    """The `PolygonLayer` renders filled, stroked and/or extruded polygons.

    !!! note

        This layer is essentially a combination of a [`PathLayer`][lonboard.PathLayer]
        and a [`SolidPolygonLayer`][lonboard.SolidPolygonLayer]. This has some overhead
        beyond a `SolidPolygonLayer`, so if you're looking for the maximum performance
        with large data, you may want to use a `SolidPolygonLayer` directly.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, PolygonLayer

    # A GeoDataFrame with Polygon or MultiPolygon geometries
    gdf = gpd.GeoDataFrame()
    layer = PolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
        get_line_color=[0, 100, 100, 150],
    )
    m = Map(layer)
    ```

    From an Arrow-compatible source like [pyogrio][pyogrio] or [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_flatgeobuf
    from lonboard import Map, PolygonLayer

    # Example: A FlatGeobuf file with Polygon or MultiPolygon geometries
    table = read_flatgeobuf("path/to/file.fgb")
    layer = PolygonLayer(
        table,
        get_fill_color=[255, 0, 0],
        get_line_color=[0, 100, 100, 150],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[PolygonLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("polygon")

    table = t.ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTIPOLYGON},
    )
    """A GeoArrow table with a Polygon or MultiPolygon column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.PolygonLayer.from_geopandas] instead.
    """

    stroked = t.Bool(None, allow_none=True)
    """Whether to draw an outline around the polygon (solid fill).

    Note that both the outer polygon as well the outlines of any holes will be drawn.

    - Type: `bool`, optional
    - Default: `True`
    """

    filled = t.Bool(None, allow_none=True)
    """Whether to draw a filled polygon (solid fill).

    Note that only the area between the outer polygon and any holes will be filled.

    - Type: `bool`, optional
    - Default: `True`
    """

    extruded = t.Bool(None, allow_none=True)
    """Whether to extrude the polygons.

    Based on the elevations provided by the `getElevation` accessor.

    If set to `false`, all polygons will be flat, this generates less geometry and is
    faster than simply returning 0 from getElevation.

    - Type: `bool`, optional
    - Default: `False`
    """

    wireframe = t.Bool(None, allow_none=True)
    """
    Whether to generate a line wireframe of the polygon. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex on the polygon.

    - Type: `bool`, optional
    - Default: `False`

    **Remarks:**

    - These lines are rendered with `GL.LINE` and will thus always be 1 pixel wide.
    - Wireframe and solid extrusions are exclusive, you'll need to create two layers
      with the same data if you want a combined rendering effect.
    """

    elevation_scale = t.Float(None, allow_none=True, min=0)
    """Elevation multiplier.

    The final elevation is calculated by `elevationScale * getElevation(d)`.
    `elevationScale` is a handy property to scale all elevation without updating the
    data.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_units = t.Unicode(None, allow_none=True)
    """
    The units of the outline width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'meters'`
    """

    line_width_scale = t.Float(None, allow_none=True, min=0)
    """
    The outline width multiplier that multiplied to all outlines of `Polygon` and
    `MultiPolygon` features if the `stroked` attribute is true.

    - Type: `float`, optional
    - Default: `1`
    """

    line_width_min_pixels = t.Float(None, allow_none=True, min=0)
    """
    The minimum outline width in pixels. This can be used to prevent the outline from
    getting too small when zoomed out.

    - Type: `float`, optional
    - Default: `0`
    """

    line_width_max_pixels = t.Float(None, allow_none=True, min=0)
    """
    The maximum outline width in pixels. This can be used to prevent the outline from
    getting too big when zoomed in.

    - Type: `float`, optional
    - Default: `None`
    """

    line_joint_rounded = t.Bool(None, allow_none=True)
    """Type of joint. If `true`, draw round joints. Otherwise draw miter joints.

    - Type: `bool`, optional
    - Default: `False`
    """

    line_miter_limit = t.Float(None, allow_none=True, min=0)
    """The maximum extent of a joint in ratio to the stroke width.

    Only works if `line_joint_rounded` is false.

    - Type: `float`, optional
    - Default: `4`
    """

    get_fill_color = t.ColorAccessor(None, allow_none=True)
    """
    The fill color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the fill color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the fill color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = t.ColorAccessor(None, allow_none=True)
    """
    The outline color of each polygon in the format of `[r, g, b, [a]]`. Each channel is
    a number between 0-255 and `a` is 255 if not supplied.

    Only applies if `stroked=True`.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the outline color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the outline color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_width = t.FloatAccessor(None, allow_none=True)
    """
    The width of the outline of each polygon, in units specified by `line_width_units`
    (default `'meters'`).

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the outline width for all polygons.
        - If an array is provided, each value in the array will be used as the outline
          width for the polygon at the same row index.
    - Default: `1`.
    """

    get_elevation = t.FloatAccessor(None, allow_none=True)
    """
    The elevation to extrude each polygon with, in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """


class SolidPolygonLayer(BaseArrowLayer):
    """The `SolidPolygonLayer` renders filled and/or extruded polygons.

    !!! note

        This layer is similar to the [`PolygonLayer`][lonboard.PolygonLayer] but will
        not render an outline around polygons. In most cases, you'll want to use the
        `PolygonLayer` directly, but for very large datasets not drawing the outline can
        significantly improve performance, in which case you may want to use this layer.

    **Example:**

    From GeoPandas:

    ```py
    import geopandas as gpd
    from lonboard import Map, SolidPolygonLayer

    # A GeoDataFrame with Polygon or MultiPolygon geometries
    gdf = gpd.GeoDataFrame()
    layer = SolidPolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```

    From an Arrow-compatible source like [pyogrio][pyogrio] or [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest):

    ```py
    from geoarrow.rust.io import read_flatgeobuf
    from lonboard import Map, SolidPolygonLayer

    # Example: A FlatGeobuf file with Polygon or MultiPolygon geometries
    table = read_flatgeobuf("path/to/file.fgb")
    layer = SolidPolygonLayer(
        table,
        get_fill_color=[255, 0, 0],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> None:
        super().__init__(table=table, _rows_per_chunk=_rows_per_chunk, **kwargs)

    @classmethod
    def from_geopandas(
        cls,
        gdf: gpd.GeoDataFrame,
        *,
        auto_downcast: bool = True,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> Self:
        return super().from_geopandas(gdf=gdf, auto_downcast=auto_downcast, **kwargs)

    @classmethod
    def from_duckdb(
        cls,
        sql: str | duckdb.DuckDBPyRelation,
        con: duckdb.DuckDBPyConnection | None = None,
        *,
        crs: str | pyproj.CRS | None = None,
        **kwargs: Unpack[SolidPolygonLayerKwargs],
    ) -> Self:
        return super().from_duckdb(sql=sql, con=con, crs=crs, **kwargs)

    _layer_type = t.Unicode("solid-polygon")

    table = t.ArrowTableTrait(
        allowed_geometry_types={EXTENSION_NAME.POLYGON, EXTENSION_NAME.MULTIPOLYGON},
    )
    """A GeoArrow table with a Polygon or MultiPolygon column.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).

    If you have a GeoPandas `GeoDataFrame`, use
    [`from_geopandas`][lonboard.SolidPolygonLayer.from_geopandas] instead.
    """

    filled = t.Bool(None, allow_none=True)
    """
    Whether to fill the polygons (based on the color provided by the
    `get_fill_color` accessor).

    - Type: `bool`, optional
    - Default: `True`
    """

    extruded = t.Bool(None, allow_none=True)
    """
    Whether to extrude the polygons (based on the elevations provided by the
    `get_elevation` accessor'). If set to `False`, all polygons will be flat, this
    generates less geometry and is faster than simply returning `0` from
    `get_elevation`.

    - Type: `bool`, optional
    - Default: `False`
    """

    wireframe = t.Bool(None, allow_none=True)
    """
    Whether to generate a line wireframe of the polygon. The outline will have
    "horizontal" lines closing the top and bottom polygons and a vertical line
    (a "strut") for each vertex on the polygon.

    - Type: `bool`, optional
    - Default: `False`

    **Remarks:**

    - These lines are rendered with `GL.LINE` and will thus always be 1 pixel wide.
    - Wireframe and solid extrusions are exclusive, you'll need to create two layers
      with the same data if you want a combined rendering effect.
    """

    elevation_scale = t.Float(None, allow_none=True, min=0)
    """
    Elevation multiplier. The final elevation is calculated by `elevation_scale *
    get_elevation(d)`. `elevation_scale` is a handy property to scale all elevation
    without updating the data.

    - Type: `float`, optional
    - Default: `1`
    """

    get_elevation = t.FloatAccessor(None, allow_none=True)
    """
    The elevation to extrude each polygon with, in meters.

    Only applies if `extruded=True`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all polygons.
        - If an array is provided, each value in the array will be used as the width for
          the polygon at the same row index.
    - Default: `1000`.
    """

    get_fill_color = t.ColorAccessor(None, allow_none=True)
    """
    The fill color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the fill color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the fill color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """

    get_line_color = t.ColorAccessor(None, allow_none=True)
    """
    The line color of each polygon in the format of `[r, g, b, [a]]`. Each channel is a
    number between 0-255 and `a` is 255 if not supplied.

    Only applies if `extruded=True`.

    - Type: [ColorAccessor][lonboard.traits.ColorAccessor], optional
        - If a single `list` or `tuple` is provided, it is used as the line color for
          all polygons.
        - If a numpy or pyarrow array is provided, each value in the array will be used
          as the line color for the polygon at the same row index.
    - Default: `[0, 0, 0, 255]`.
    """



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy as np

import lonboard.traits as t
from lonboard._geoarrow.ops import Bbox, WeightedCentroid
from lonboard._utils import auto_downcast as _auto_downcast
from lonboard.layer._polygon import PolygonLayer

if TYPE_CHECKING:
    import sys
    from typing import Self

    import pandas as pd
    from arro3.core import ChunkedArray
    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import H3AccessorInput, H3HexagonLayerKwargs

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack

SEED_VALUE = 42
RNG = np.random.default_rng(SEED_VALUE)
MAX_SAMPLE_N = 10000


def default_h3_viewport(ca: ChunkedArray) -> tuple[Bbox, WeightedCentroid] | None:
    try:
        import h3.api.numpy_int as h3
    except ImportError:
        warnings.warn(
            "h3-py is not installed, cannot compute default H3 viewport.",
            ImportWarning,
        )
        return None

    sample_n = min(MAX_SAMPLE_N, len(ca))
    sample_h3 = RNG.choice(ca.to_numpy(), size=sample_n, replace=False)

    if not hasattr(h3, "cells_to_geo"):
        warnings.warn(
            "h3-py v4 is not installed, cannot compute default H3 viewport.",
            ImportWarning,
        )
        return None

    geo_dict = h3.cells_to_geo(sample_h3)

    geom_type = geo_dict.get("type")
    if geom_type == "Polygon":
        polygons = [geo_dict["coordinates"]]
    elif geom_type == "MultiPolygon":
        polygons = geo_dict["coordinates"]
    else:
        # I think it should always be Polygon/MultiPolygon
        return None

    coords = np.array(
        [pt for polygon in polygons for ring in polygon for pt in ring],
        dtype=np.float64,
    )
    centroid = coords.mean(axis=0)

    minx, miny = coords.min(axis=0)
    maxx, maxy = coords.max(axis=0)

    return (
        Bbox(minx=float(minx), miny=float(miny), maxx=float(maxx), maxy=float(maxy)),
        # Still give it the weight of the full input dataset
        WeightedCentroid(x=float(centroid[0]), y=float(centroid[1]), num_items=len(ca)),
    )


class H3HexagonLayer(PolygonLayer):
    """The `H3HexagonLayer` renders hexagons from the [H3](https://h3geo.org/) geospatial indexing system.

    **Example:**

    From Pandas:

    ```py
    import pandas as pd
    from lonboard import Map, H3HexagonLayer

    # A DataFrame with H3 cell identifiers
    df = pd.DataFrame({
        "h3_index": ["8928308280fffff", "8928308280bffff", ...],
        "other_attributes": [...],
    })
    layer = H3HexagonLayer.from_pandas(
        df,
        get_hexagon=df["h3_index"],
    )
    m = Map(layer)
    ```

    Or, you can pass in an Arrow table directly

    ```py
    from lonboard import Map, H3HexagonLayer

    # Example: An Arrow table with H3 identifiers as a column
    layer = H3HexagonLayer(
        table,
        get_hexagon=table["h3_index"],
    )
    m = Map(layer)
    ```
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        get_hexagon: H3AccessorInput,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[H3HexagonLayerKwargs],
    ) -> None:
        """Create a new H3HexagonLayer.

        Args:
            table: An Arrow table with properties to associate with the H3 hexagons.

        Keyword Args:
            get_hexagon: The cell identifier of each H3 hexagon.
            kwargs: Extra args passed down as H3HexagonLayer attributes.

        """
        super().__init__(
            table=table,
            get_hexagon=get_hexagon,
            _rows_per_chunk=_rows_per_chunk,
            **kwargs,
        )

        # Assign viewport after get_hexagon has already been validated to be uint64
        # array
        default_viewport = default_h3_viewport(self.get_hexagon)
        if default_viewport is not None:
            self._bbox = default_viewport[0]
            self._weighted_centroid = default_viewport[1]

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        *,
        get_hexagon: H3AccessorInput,
        auto_downcast: bool = True,
        **kwargs: Unpack[H3HexagonLayerKwargs],
    ) -> Self:
        """Create a new H3HexagonLayer from a pandas DataFrame.

        Args:
            df: a Pandas DataFrame with properties to associate with H3 hexagons.

        Keyword Args:
            get_hexagon: H3 cell identifier of each H3 hexagon.
            auto_downcast: Whether to save memory on input by casting to smaller types. Defaults to True.
            kwargs: Extra args passed down as H3HexagonLayer attributes.

        """
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow required for converting GeoPandas to arrow.\n"
                "Run `pip install pyarrow`.",
            ) from e

        if auto_downcast:
            # Note: we don't deep copy because we don't need to clone geometries
            df = _auto_downcast(df.copy())  # type: ignore

        table = pa.Table.from_pandas(df)
        return cls(table, get_hexagon=get_hexagon, **kwargs)

    _layer_type = t.Unicode("h3-hexagon")

    table = t.ArrowTableTrait(geometry_required=False)
    """An Arrow table with properties to associate with the H3 hexagons.

    If you have a Pandas `DataFrame`, use
    [`from_pandas`][lonboard.H3HexagonLayer.from_pandas] instead.
    """

    get_hexagon = t.H3Accessor()
    """The cell identifier of each H3 hexagon.

    Accepts either an array of strings or uint64 integers representing H3 cell IDs.

    - Type: [H3Accessor][lonboard.traits.H3Accessor]
    """

    high_precision = t.Bool(None, allow_none=True)
    """Whether to render H3 hexagons in high-precision mode.

    Each hexagon in the H3 indexing system is [slightly different in shape](https://h3geo.org/docs/core-library/coordsystems). To draw a large number of hexagons efficiently, the `H3HexagonLayer` may choose to use instanced drawing by assuming that all hexagons within the current viewport have the same shape as the one at the center of the current viewport. The discrepancy is usually too small to be visible.

    There are several cases in which high-precision mode is required. In these cases, `H3HexagonLayer` may choose to switch to high-precision mode, where it trades performance for accuracy:

    * The input set contains a pentagon. There are 12 pentagons world wide at each resolution, and these cells and their immediate neighbors have significant differences in shape.
    * The input set is at a coarse resolution (res `0` through res `5`). These cells have larger differences in shape, particularly when using a Mercator projection.
    * The input set contains hexagons with different resolutions.

    Possible values:

    * `None`: The layer chooses the mode automatically. High-precision rendering is only used if an edge case is encountered in the data.
    * `True`: Always use high-precision rendering.
    * `False`: Always use instanced rendering, regardless of the characteristics of the data.

    - Type: `bool | None`, optional
    - Default: `None`
    """

    coverage = t.Float(None, allow_none=True, min=0, max=1)
    """Hexagon radius multiplier, between 0 - 1.

    When coverage = 1, hexagon is rendered with actual size, by specifying a different value (between 0 and 1) hexagon can be scaled down.

    - Type: `float`, optional
    - Default: `1`
    """



from typing import Any as TypingAny

import traitlets.traitlets as t


class Any(t.Any):
    """A subclass of [Any][traitlets.Any] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Bool(t.Bool):
    """A subclass of [Bool][traitlets.Bool] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Enum(t.Enum):
    """A subclass of [Enum][traitlets.Enum] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Float(t.Float):
    """A subclass of [Float][traitlets.Float] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Int(t.Int):
    """A subclass of [Int][traitlets.Int] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class List(t.List):
    """A subclass of [List][traitlets.List] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Tuple(t.Tuple):
    """A subclass of [Tuple][traitlets.Tuple] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Unicode(t.Unicode):
    """A subclass of [Unicode][traitlets.Unicode] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)


class Union(t.Union):
    """A subclass of [Union][traitlets.Union] tagged with `sync=True`."""

    def __init__(self, *args: TypingAny, **kwargs: TypingAny) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)



"""Python library for fast, interactive geospatial data visualization in Jupyter."""

from . import basemap, colormap, controls, experimental, layer_extension, traits
from ._map import Map
from ._version import __version__
from ._viz import viz
from .layer import (
    A5Layer,
    ArcLayer,
    BaseArrowLayer,
    BaseLayer,
    BitmapLayer,
    BitmapTileLayer,
    ColumnLayer,
    GeohashLayer,
    H3HexagonLayer,
    HeatmapLayer,
    PathLayer,
    PointCloudLayer,
    PolygonLayer,
    RasterLayer,
    S2Layer,
    ScatterplotLayer,
    SolidPolygonLayer,
    TripsLayer,
)

__all__ = [
    "A5Layer",
    "ArcLayer",
    "BaseArrowLayer",
    "BaseLayer",
    "BitmapLayer",
    "BitmapTileLayer",
    "ColumnLayer",
    "GeohashLayer",
    "H3HexagonLayer",
    "HeatmapLayer",
    "Map",
    "PathLayer",
    "PointCloudLayer",
    "PolygonLayer",
    "RasterLayer",
    "S2Layer",
    "ScatterplotLayer",
    "SolidPolygonLayer",
    "TripsLayer",
    "__version__",
    "basemap",
    "colormap",
    "controls",
    "experimental",
    "layer_extension",
    "traits",
    "viz",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pyproj
from affine import Affine
from arro3.core import DataType

import lonboard.traits as t
from lonboard.experimental.traits import MeshAccessor, TextureTrait
from lonboard.layer import BaseLayer

if TYPE_CHECKING:
    from typing import Any, Self

    from numpy.typing import NDArray
    from rasterio.io import DatasetReader


def load_arr_and_transform(
    src: DatasetReader,
    *,
    downscale: int | None,
) -> tuple[NDArray[np.uint8], Affine]:
    """Load array and transform from rasterio source."""
    if downscale is None:
        return src.read(), src.transform

    # Read overview array from src
    overview_height = int(src.height / downscale)
    overview_width = int(src.width / downscale)
    overview_shape = (src.count, overview_height, overview_width)

    arr: np.ndarray = src.read(out_shape=overview_shape)
    overview_transform = Affine(
        src.transform.a * downscale,  # pixel width
        src.transform.b,  # rotation/skew x
        src.transform.c,  # top-left x
        src.transform.d,  # rotation/skew y
        src.transform.e * downscale,  # pixel height (usually negative)
        src.transform.f,  # top-left y
    )
    return arr, overview_transform


def apply_colormap(
    arr: NDArray[np.uint8],
    cmap: dict[int, tuple[int, int, int] | tuple[int, int, int, int]],
) -> NDArray[np.uint8]:
    """Apply rasterio colormap to single-band array."""
    lut = np.zeros((max(cmap.keys()) + 1, 4), dtype=np.uint8)
    for k, v in cmap.items():
        lut[k] = v

    return lut[arr]


def generate_mesh_grid(
    *,
    n_rows: int = 50,
    n_cols: int = 50,
) -> tuple[NDArray[np.float32], NDArray[np.uint32]]:
    """Generate a regular grid mesh with the given number of rows and columns.

    Creates a mesh covering the unit square [0, 1] x [0, 1] divided into
    n_rows x n_cols cells. Each cell is split into two triangles along the diagonal.

    Args:
        n_rows: Number of rows in the grid
        n_cols: Number of columns in the grid

    Returns:
        positions: Array of shape ((n_rows+1) * (n_cols+1), 2) containing vertex positions
                   in normalized image coordinates [0, 1]
        triangles: Array of shape (n_rows * n_cols * 2, 3) containing triangle vertex indices

    """
    # Generate vertex grid
    # We need (n_rows + 1) x (n_cols + 1) vertices to create n_rows x n_cols cells
    x = np.linspace(0, 1, n_cols + 1, dtype=np.float32)
    y = np.linspace(0, 1, n_rows + 1, dtype=np.float32)

    # Create meshgrid and flatten to get all vertex positions
    xx, yy = np.meshgrid(x, y)
    positions = np.stack([xx.ravel(), yy.ravel()], axis=-1)

    # Generate triangle indices
    # For each cell (i, j), we create two triangles:
    # Triangle 1: bottom-left, bottom-right, top-left
    # Triangle 2: bottom-right, top-right, top-left
    triangles = np.empty((n_rows * n_cols * 2, 3), dtype=np.uint32)

    i = 0
    for row in range(n_rows):
        for col in range(n_cols):
            # Vertex indices for the current cell
            bottom_left = row * (n_cols + 1) + col
            bottom_right = bottom_left + 1
            top_left = (row + 1) * (n_cols + 1) + col
            top_right = top_left + 1

            triangles[i] = [bottom_left, bottom_right, top_left]
            triangles[i + 1] = [bottom_right, top_right, top_left]
            i += 2

    triangles_array = np.array(triangles, dtype=np.uint32)

    return positions, triangles_array


def rescale_positions_to_image_crs(
    positions: NDArray[np.float32],
    *,
    image_height: int,
    image_width: int,
    transform: Affine,
) -> NDArray[np.float32]:
    # Scale to pixel coordinates
    pixel_coords = positions * np.array([image_width, image_height])
    cols, rows = pixel_coords[:, 0], pixel_coords[:, 1]
    xs, ys = transform * (cols, rows)  # type: ignore (matrix multiplication not typed correctly)
    return np.stack([xs, ys], axis=-1)


class SurfaceLayer(BaseLayer):
    """SurfaceLayer."""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)  # type: ignore

    @classmethod
    def from_rasterio(
        cls,
        src: DatasetReader,
        *,
        downscale: int | None = None,
        mesh_n_rows: int = 50,
        mesh_n_cols: int = 50,
        **kwargs: Any,
    ) -> Self:
        import rasterio.plot

        arr, transform = load_arr_and_transform(src, downscale=downscale)

        if arr.shape[0] == 1 and src.colormap(1) is not None:
            image_arr = apply_colormap(arr[0], src.colormap(1))
        else:
            # swap axes order from (bands, rows, columns) to (rows, columns, bands)
            image_arr = rasterio.plot.reshape_as_image(arr)

        image_height = image_arr.shape[0]
        image_width = image_arr.shape[1]

        # Generate mesh grid in normalized [0, 1] coordinates
        # These positions serve as texture coordinates
        tex_coords, triangles = generate_mesh_grid(
            n_rows=mesh_n_rows,
            n_cols=mesh_n_cols,
        )

        # Reproject mesh vertices from image CRS to EPSG:4326
        source_crs_coords = rescale_positions_to_image_crs(
            tex_coords,
            image_height=image_height,
            image_width=image_width,
            transform=transform,
        )

        transformer = pyproj.Transformer.from_crs(src.crs, "EPSG:4326", always_xy=True)

        lons, lats = transformer.transform(
            source_crs_coords[:, 0],
            source_crs_coords[:, 1],
        )
        lonlat_coords = np.stack([lons, lats], axis=-1)

        # Add z-coordinate (elevation), setting to zero as we care about a flat mesh on
        # the map surface
        final_positions = np.concatenate(
            [lonlat_coords, np.zeros((lonlat_coords.shape[0], 1), dtype=np.float32)],
            axis=-1,
        )

        return cls(
            positions=final_positions,
            triangles=triangles,
            tex_coords=tex_coords,
            texture=image_arr,
            **kwargs,
        )

    _layer_type = t.Unicode("surface")

    positions = MeshAccessor(
        list_size=3,
        expected_arrow_type=DataType.list(
            DataType.float32(),
            3,
        ),
    )
    tex_coords = MeshAccessor(
        list_size=2,
        expected_arrow_type=DataType.list(
            DataType.float32(),
            2,
        ),
    )
    triangles = MeshAccessor(
        list_size=3,
        expected_arrow_type=DataType.list(
            DataType.uint32(),
            3,
        ),
    )

    texture = TextureTrait()

    wireframe = t.Bool(None, allow_none=True)
    """Whether to render the mesh in wireframe mode.

    - Type: `bool`, optional
    - Default: `False`
    """



from __future__ import annotations

from typing import TYPE_CHECKING

import lonboard.traits as t
from lonboard._utils import auto_downcast as _auto_downcast

# Important to import from ._polygon to avoid circular imports
from lonboard.layer._polygon import PolygonLayer

if TYPE_CHECKING:
    import sys
    from typing import Self

    import pandas as pd
    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import GeohashLayerKwargs, TextAccessorInput

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class GeohashLayer(PolygonLayer):
    """The `GeohashLayer` renders filled and/or stroked polygons based on the [Geohash](https://en.wikipedia.org/wiki/Geohash) geospatial indexing system.

    !!! warning
        This layer does not currently support auto-centering the map view based on the
        extent of the input.
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        get_geohash: TextAccessorInput,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[GeohashLayerKwargs],
    ) -> None:
        """Create a new GeohashLayer.

        Args:
            table: An Arrow table with properties to associate with the geohashes.

        Keyword Args:
            get_geohash: The identifier of each geohash.
            kwargs: Extra args passed down as GeohashLayer attributes.

        """
        super().__init__(
            table=table,
            get_geohash=get_geohash,
            _rows_per_chunk=_rows_per_chunk,
            **kwargs,
        )

    @classmethod
    def from_pandas(
        cls,
        df: pd.DataFrame,
        *,
        get_geohash: TextAccessorInput,
        auto_downcast: bool = True,
        **kwargs: Unpack[GeohashLayerKwargs],
    ) -> Self:
        """Create a new GeohashLayer from a pandas DataFrame.

        Args:
            df: a Pandas DataFrame with properties to associate with geohashes.

        Keyword Args:
            get_geohash: geohash identifiers.
            auto_downcast: Whether to save memory on input by casting to smaller types. Defaults to True.
            kwargs: Extra args passed down as GeohashLayer attributes.

        """
        try:
            import pyarrow as pa
        except ImportError as e:
            raise ImportError(
                "pyarrow required for converting GeoPandas to arrow.\n"
                "Run `pip install pyarrow`.",
            ) from e

        if auto_downcast:
            # Note: we don't deep copy because we don't need to clone geometries
            df = _auto_downcast(df.copy())  # type: ignore

        table = pa.Table.from_pandas(df)
        return cls(table, get_geohash=get_geohash, **kwargs)

    _layer_type = t.Unicode("geohash")

    table = t.ArrowTableTrait(geometry_required=False)
    """An Arrow table with properties to associate with the geohashes.

    If you have a Pandas `DataFrame`, use
    [`from_pandas`][lonboard.GeohashLayer.from_pandas] instead.
    """

    get_geohash = t.TextAccessor()
    """The cell identifier of each geohash.

    Accepts either an array of strings or uint64 integers representing geohash IDs.

    - Type: [TextAccessor][lonboard.traits.TextAccessor]
    """



from __future__ import annotations

import lonboard.traits as t
from lonboard.controls._base import BaseControl


class ScaleControl(BaseControl):
    """A deck.gl ScaleControl.

    Passing this to [`Map.controls`][lonboard.Map.controls] will add a scale bar to the
    map.
    """

    _control_type = t.Unicode("scale")

    max_width = t.Int(allow_none=True, default_value=None)
    """The maximum width of the scale control in pixels.

    This only has effect for Maplibre-driven maps (i.e. where
    [`MaplibreBasemap.mode`][lonboard.basemap.MaplibreBasemap.mode] is "overlaid" or
    "interleaved").

    Default `100`.
    """

    unit = t.Unicode(allow_none=True, default_value=None)
    """The unit of the scale.

    This only has effect for Maplibre-driven maps (i.e. where
    [`MaplibreBasemap.mode`][lonboard.basemap.MaplibreBasemap.mode] is "overlaid" or
    "interleaved").

    One of `'metric'`, `'imperial'`, or `'nautical'`. Default is `'metric'`.
    """



"""Basemap helpers."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING

from typing_extensions import deprecated

import lonboard.traits as t
from lonboard._base import BaseWidget
from lonboard.traits import BasemapUrl

if TYPE_CHECKING:
    from typing import Literal


class CartoStyle(StrEnum):
    """Maplibre-supported vector basemap styles provided by Carto.

    Refer to [Carto
    documentation](https://docs.carto.com/carto-for-developers/carto-for-react/guides/basemaps)
    for information on styles.
    """

    DarkMatter = "https://basemaps.cartocdn.com/gl/dark-matter-gl-style/style.json"
    """A dark map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/dark_labels.png)
    """

    DarkMatterNoLabels = (
        "https://basemaps.cartocdn.com/gl/dark-matter-nolabels-gl-style/style.json"
    )
    """A dark map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/dark_no_labels.png)
    """

    Positron = "https://basemaps.cartocdn.com/gl/positron-gl-style/style.json"
    """A light map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/positron_labels.png)
    """

    PositronNoLabels = (
        "https://basemaps.cartocdn.com/gl/positron-nolabels-gl-style/style.json"
    )
    """A light map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/positron_no_labels.png)
    """

    Voyager = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json"
    """A light, colored map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/voyager_labels.png)
    """

    VoyagerNoLabels = (
        "https://basemaps.cartocdn.com/gl/voyager-nolabels-gl-style/style.json"
    )
    """A light, colored map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/voyager_no_labels.png)
    """


# NOTE: this is fully duplicated because you can't subclass enums, and I don't see any
# other way to provide a deprecation warning
@deprecated(
    "CartoBasemap is deprecated, use CartoStyle instead. Will be removed in v0.14",
)
class CartoBasemap(StrEnum):
    """Maplibre-supported vector basemap styles provided by Carto.

    Refer to [Carto
    documentation](https://docs.carto.com/carto-for-developers/carto-for-react/guides/basemaps)
    for information on styles.
    """

    DarkMatter = "https://basemaps.cartocdn.com/gl/dark-matter-gl-style/style.json"
    """A dark map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/dark_labels.png)
    """

    DarkMatterNoLabels = (
        "https://basemaps.cartocdn.com/gl/dark-matter-nolabels-gl-style/style.json"
    )
    """A dark map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/dark_no_labels.png)
    """

    Positron = "https://basemaps.cartocdn.com/gl/positron-gl-style/style.json"
    """A light map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/positron_labels.png)
    """

    PositronNoLabels = (
        "https://basemaps.cartocdn.com/gl/positron-nolabels-gl-style/style.json"
    )
    """A light map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/positron_no_labels.png)
    """

    Voyager = "https://basemaps.cartocdn.com/gl/voyager-gl-style/style.json"
    """A light, colored map style with labels.

    ![](https://carto.com/help/images/building-maps/basemaps/voyager_labels.png)
    """

    VoyagerNoLabels = (
        "https://basemaps.cartocdn.com/gl/voyager-nolabels-gl-style/style.json"
    )
    """A light, colored map style without labels.

    ![](https://carto.com/help/images/building-maps/basemaps/voyager_no_labels.png)
    """


class MaplibreBasemap(BaseWidget):
    """A [MapLibre GL JS](https://maplibre.org/maplibre-gl-js/docs/) basemap."""

    def __init__(
        self,
        *,
        mode: Literal[
            "interleaved",
            "overlaid",
            "reverse-controlled",
        ] = "overlaid",
        style: str | CartoStyle = CartoStyle.PositronNoLabels,
    ) -> None:
        """Create a MapLibre GL JS basemap."""
        super().__init__(mode=mode, style=style)

    mode = t.Enum(
        [
            "interleaved",
            "overlaid",
            "reverse-controlled",
        ],
        default_value="overlaid",
    )
    """The basemap integration mode.

    This determines how deck.gl and MapLibre are rendered together.

    - **`"interleaved"`**:

        The interleaved mode renders deck.gl layers into the same context created by
        MapLibre.

        If you need to mix deck.gl layers with MapLibre layers, e.g. having deck.gl
        surfaces below text labels, or objects occluding each other correctly in 3D,
        then you have to use this option.

        See [BaseLayer.before_id][lonboard.BaseLayer.before_id] for more information on
        how to interleave deck.gl layers into the Maplibre layer stack.

    - **`"overlaid"`**:

        The overlaid mode renders deck.gl in a separate canvas inside the MapLibre's controls container.

        If your use case does not require interleaving, but you still want to use
        certain features of maplibre-gl, such as Maplibre-based controls, then you
        should use this option.

    - **`"reverse-controlled"`**:

        The reverse-controlled mode renders deck.gl above the MapLibre container and blocks any interaction to the base map.

        This is the mode that Lonboard has historically used.

    **Default**: `"overlaid"`
    """

    style = BasemapUrl(CartoStyle.PositronNoLabels)
    """
    A MapLibre-compatible basemap style.

    Various styles are provided in [`CartoStyle`][lonboard.basemap.CartoStyle].

    - Type: `str`, holding a URL hosting a basemap style.
    - Default: [`CartoStyle.PositronNoLabels`][lonboard.basemap.CartoStyle.PositronNoLabels]
    """



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def str_to_h3(hex_arr: NDArray[np.str_]) -> NDArray[np.uint64]:
    """Convert an array of hexadecimal strings to H3 indices (uint64).

    This is a pure NumPy vectorized implementation that processes hex strings
    character by character without Python loops.

    Args:
        hex_arr: Array of hexadecimal strings (15 characters each)

    Returns:
        Array of H3 indices as uint64 integers

    """
    if len(hex_arr) == 0:
        return np.array([], dtype=np.uint64)

    # Convert to S15 fixed-width byte strings if needed
    # View as 2D array of individual bytes (shape: n x 15)
    hex_bytes = np.asarray(hex_arr, dtype="S15").view("S1").reshape(len(hex_arr), -1)

    # Convert ASCII bytes to numeric values
    # Get the ASCII code of each character
    ascii_vals = hex_bytes.view(np.uint8)

    # Convert hex ASCII to numeric values (0-15)
    # '0'-'9' (48-57) -> 0-9
    # 'A'-'F' (65-70) -> 10-15
    # 'a'-'f' (97-102) -> 10-15
    vals = ascii_vals - ord("0")  # Shift '0' to 0
    vals = np.where(vals > 9, vals - 7, vals)  # 'A'=65-48=17 -> 17-7=10
    vals = np.where(vals > 15, vals - 32, vals)  # 'a'=97-48=49 -> 49-7=42 -> 42-32=10

    # Create powers of 16 for each position (most significant first)
    # For 15 hex digits: [16^14, 16^13, ..., 16^1, 16^0]
    n_digits = hex_bytes.shape[1]
    powers = 16 ** np.arange(n_digits - 1, -1, -1, dtype=np.uint64)

    # Compute dot product to get final uint64 values
    # Each row: sum(digit_i * 16^(n-1-i))
    return np.dot(vals.astype(np.uint64), powers)



"""Manage the display of exceptions raised by JavaScript code."""

from ipywidgets import Output


class ErrorOutput(Output):
    """An Output widget that captures and displays error messages."""

    name: str | None

    def __init__(self, *, name: str | None = None) -> None:
        self.name = name

        super().__init__()

    def _prepare_text(self, text: str) -> str:
        if self.name is not None:
            return f"[{self.name}] {text}\n"

        return text + "\n"

    def append_stdout(self, text: str) -> None:
        return super().append_stdout(self._prepare_text(text))

    def append_stderr(self, text: str) -> None:
        return super().append_stderr(self._prepare_text(text))



# ruff: noqa: C901, PLR0912, SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from arro3.core import Array, ChunkedArray, DataType, fixed_size_list_array

from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard._vendor.matplotlib.colors import _to_rgba_no_colorcycle
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    from lonboard.layer import BaseArrowLayer


class ColorAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl color accessor.

    Various input is allowed:

    - A `list` or `tuple` with three or four integers, ranging between 0 and 255
      (inclusive). This will be used as the color for all objects.
    - A `str` representing a hex color or "well known" color interpretable by
      [matplotlib.colors.to_rgba][matplotlib.colors.to_rgba]. See also matplotlib's
      [list of named
      colors](https://matplotlib.org/stable/gallery/color/named_colors.html#base-colors).
    - A numpy `ndarray` with two dimensions and data type [`np.uint8`][numpy.uint8]. The
      size of the second dimension must be `3` or `4`, and will correspond to either RGB
      or RGBA colors.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `3` or `4` and its child must have type
      `uint8`.
    - Any Arrow fixed size list array from a library that implements the [Arrow
      PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).

    You can use helpers in the [`lonboard.colormap`][lonboard.colormap] module (i.e.
    [`apply_continuous_cmap`][lonboard.colormap.apply_continuous_cmap]) to simplify
    constructing numpy arrays for color values.
    """

    default_value = (0, 0, 0)
    info_text = (
        "a tuple or list representing an RGB(A) color or numpy ndarray or "
        "pyarrow FixedSizeList representing an array of RGB(A) colors"
    )

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if not np.issubdtype(value.dtype, np.uint8):
            self.error(obj, value, info="Color array must be uint8 type.")

        if value.ndim != 2:
            self.error(obj, value, info="Color array must have 2 dimensions.")

        list_size = value.shape[1]
        if list_size not in (3, 4):
            self.error(
                obj,
                value,
                info="Color array must have 3 or 4 as its second dimension.",
            )

        return ChunkedArray([fixed_size_list_array(value.ravel("C"), list_size)])

    def validate(self, obj: BaseArrowLayer, value: Any) -> tuple | list | ChunkedArray:
        if isinstance(value, (tuple, list)):
            if len(value) < 3 or len(value) > 4:
                self.error(obj, value, info="3 or 4 values if passed a tuple or list")

            if any(not isinstance(v, int) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values to be integers if passed a tuple or list",
                )

            if any(v < 0 or v > 255 for v in value):
                self.error(
                    obj,
                    value,
                    info="values between 0 and 255",
                )

            return value

        if isinstance(value, str):
            try:
                c = _to_rgba_no_colorcycle(value)  # type: ignore
            except ValueError:
                return self.error(
                    obj,
                    value,
                    info=(
                        "Color string must be a named color or hex string interpretable"
                        " by matplotlib.colors.to_rgba."
                    ),
                )

            return tuple(map(int, (np.array(c) * 255).astype(np.uint8)))

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="Color Arrow array must be a FixedSizeList.")

        if value.type.list_size not in (3, 4):
            self.error(
                obj,
                value,
                info=(
                    "Color Arrow array must have a FixedSizeList inner size of 3 or 4."
                ),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_uint8(value_type):
            self.error(obj, value, info="Color Arrow array must have a uint8 child.")

        return value.rechunk(max_chunksize=obj._rows_per_chunk)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def h3_to_str(h3_indices: NDArray[np.uint64]) -> NDArray[np.str_]:
    """Convert an array of H3 indices (uint64) to their hexadecimal string representations.

    Returns a numpy array of type S15 (fixed-length ASCII strings of length 15).
    """
    # Ensure input is a numpy array of uint64
    hex_chars = np.empty((h3_indices.size, 15), dtype="S1")

    # Prepare hex digits lookup
    hex_digits = np.array(list("0123456789ABCDEF"), dtype="S1")

    # Fill each digit
    for i in range(15):
        shift = (15 - 1 - i) * 4
        hex_chars[:, i] = hex_digits[(h3_indices >> shift) & 0xF]

    return hex_chars.view("<S15")[:, 0]



# ruff: noqa: SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from arro3.core import Array, ChunkedArray, DataType, Field, fixed_size_list_array

from lonboard._geoarrow.extension_types import CoordinateDimension, coord_storage_type
from lonboard._geoarrow.ops.coord_layout import convert_struct_column_to_interleaved
from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    from lonboard.layer import BaseArrowLayer


class PointAccessor(FixedErrorTraitType):
    """A representation of a deck.gl point accessor.

    Various input is allowed:

    - A numpy `ndarray` with two dimensions and data type [`np.float64`][numpy.float64].
      The size of the second dimension must be `2` or `3`, and will correspond to either
      XY or XYZ positions.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray],
      [`StructArray`][pyarrow.StructArray] or [`ChunkedArray`][pyarrow.ChunkedArray]
      containing `FixedSizeListArray`s or `StructArray`s. The inner size of the fixed
      size list must be `2` or `3` and its child must be of floating point type.
    - Any Arrow fixed size list or struct array from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = (0, 0, 0)
    info_text = "a numpy ndarray, arrow FixedSizeList, or arrow Struct representing a point array"

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if value.ndim != 2:
            self.error(obj, value, info="Point array to have 2 dimensions")

        list_size = value.shape[1]
        if list_size not in (2, 3):
            self.error(
                obj,
                value,
                info="Point array to have 2 or 3 as its second dimension",
            )

        if not np.issubdtype(value.dtype, np.float64):
            self.error(obj, value, info="Point array to have float64 type.")

        # Set geoarrow extension metadata
        field = Field(
            "",
            coord_storage_type(
                interleaved=True,
                dims=CoordinateDimension.XY
                if list_size == 2
                else CoordinateDimension.XYZ,
            ),
            nullable=True,
            metadata={"ARROW:extension:name": "geoarrow.point"},
        )
        array = fixed_size_list_array(
            value.ravel("C"),
            list_size,
            type=field,
        )

        return ChunkedArray([array])

    def validate(
        self,
        obj: BaseArrowLayer,
        value: Any,
    ) -> tuple[int, ...] | list[int] | ChunkedArray:
        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)
        _, value = convert_struct_column_to_interleaved(field=value.field, column=value)

        if not DataType.is_fixed_size_list(value.type):
            self.error(obj, value, info="Point arrow array to be a FixedSizeList")

        if value.type.list_size not in (2, 3):
            self.error(
                obj,
                value,
                info=(
                    "point arrow array to be a FixedSizeList with list size of 2 or 3"
                ),
            )

        value_type = value.type.value_type
        assert value_type is not None
        if not DataType.is_float64(value_type):
            self.error(
                obj,
                value,
                info="Point arrow array to have a float64 point child",
            )

        return value.rechunk(max_chunksize=obj._rows_per_chunk)



# ruff: noqa: SLF001

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from arro3.core import Array, ChunkedArray, DataType

from lonboard._h3._str_to_h3 import str_to_h3
from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard.traits._base import FixedErrorTraitType

if TYPE_CHECKING:
    import pandas as pd
    from numpy.typing import NDArray

    from lonboard.layer import BaseArrowLayer


class A5Accessor(FixedErrorTraitType):
    """A trait to validate A5 cell input.

    Various input is allowed:

    - A numpy `ndarray` with an object, S16, or uint64 data type.
    - A pandas `Series` with an object or uint64 data type.
    - A pyarrow string, large string, string view array, or uint64 array, or a chunked array of those types.
    - Any Arrow string, large string, string view array, or uint64 array, or a chunked array of those types from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = None
    info_text = (
        "a float value or numpy ndarray or Arrow array representing an array of floats"
    )

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def _pandas_to_numpy(
        self,
        obj: BaseArrowLayer,
        value: pd.Series,
    ) -> NDArray[np.str_] | NDArray[np.uint64]:
        """Cast pandas Series to numpy ndarray."""
        if isinstance(value.dtype, np.dtype) and np.issubdtype(value.dtype, np.integer):
            return np.asarray(value, dtype=np.uint64)

        if not isinstance(value.dtype, np.dtype) or not np.issubdtype(
            value.dtype,
            np.object_,
        ):
            self.error(
                obj,
                value,
                info="A5 Pandas series not object or uint64 dtype.",
            )

        if not (value.str.len() == 16).all():
            self.error(
                obj,
                value,
                info="A5 Pandas series not all 16 characters long.",
            )

        return np.asarray(value, dtype="S16")

    def _numpy_to_arrow(self, obj: BaseArrowLayer, value: np.ndarray) -> ChunkedArray:
        if np.issubdtype(value.dtype, np.uint64):
            return ChunkedArray([value])

        if np.issubdtype(value.dtype, np.object_):
            if {len(v) for v in value} != {16}:
                self.error(
                    obj,
                    value,
                    info="numpy object array not all 16 characters long",
                )

            value = np.asarray(value, dtype="S16")

        if not np.issubdtype(value.dtype, np.dtype("S16")):
            self.error(obj, value, info="numpy array not object, str, or uint64 dtype")

        a5_uint8_array = str_to_h3(value)
        return ChunkedArray([a5_uint8_array])

    def validate(self, obj: BaseArrowLayer, value: Any) -> ChunkedArray:
        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            value = self._pandas_to_numpy(obj, value)

        if isinstance(value, np.ndarray):
            value = self._numpy_to_arrow(obj, value)
        elif hasattr(value, "__arrow_c_array__"):
            value = ChunkedArray([Array.from_arrow(value)])
        elif hasattr(value, "__arrow_c_stream__"):
            value = ChunkedArray.from_arrow(value)
        else:
            self.error(obj, value)

        assert isinstance(value, ChunkedArray)

        if (
            DataType.is_string(value.type)
            or DataType.is_large_string(value.type)
            or DataType.is_string_view(value.type)
        ):
            value = self._numpy_to_arrow(obj, value.to_numpy())

        if not DataType.is_uint64(value.type):
            self.error(
                obj,
                value,
                info="A5 Arrow array must be uint64 type.",
            )

        return value.rechunk(max_chunksize=obj._rows_per_chunk)



# ruff: noqa: SLF001

from __future__ import annotations

import asyncio
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    Literal,
    NotRequired,
    Protocol,
    TypedDict,
    cast,
)

import lonboard.traits as t
from lonboard.controls._base import BaseControl

if TYPE_CHECKING:
    from collections.abc import Sequence

    from geopy.geocoders.base import Geocoder


GEOCODER_MSG_KIND: Final = "geocoder-query"


class _GeocoderRequestMessage(TypedDict):
    query: str


class _GeocoderRequest(TypedDict):
    id: str
    kind: Literal["geocoder-query"]
    msg: _GeocoderRequestMessage


def _on_geocoder_dispatch(
    widget: GeocoderControl,
    msg: str | list | dict,
    buffers: list[bytes],  # noqa: ARG001
) -> None:
    """Synchronous dispatch entry point passed to `Widget.on_msg`."""  # noqa: D401
    if not isinstance(msg, dict) or msg.get("kind") != GEOCODER_MSG_KIND:
        return

    # Schedule the async handler on the event loop
    task = asyncio.create_task(
        _handle_geocoder_request(widget, cast("_GeocoderRequest", msg)),
    )
    widget._pending_tasks.add(task)
    task.add_done_callback(widget._pending_tasks.discard)


async def _handle_geocoder_request(
    widget: GeocoderControl,
    msg: _GeocoderRequest,
) -> None:
    """Async callback to handle geocoder requests from the frontend."""
    output = widget._error_output

    try:
        if widget.debug:
            output.append_stdout(f"Received geocoder request: {msg}")

        query = msg["msg"]["query"]
        response = await widget._client(query)
        widget.send(
            {
                "id": msg["id"],
                "kind": f"{GEOCODER_MSG_KIND}-response",
                "response": response,
            },
        )

    except Exception:  # noqa: BLE001
        error_msg = traceback.format_exc()
        output.append_stderr(f"Error performing geocoder query: {error_msg}\n")

        widget.send(
            {
                "id": msg["id"],
                "kind": f"{GEOCODER_MSG_KIND}-response",
                "response": {"error": error_msg},
            },
        )


class GeoJsonPoint(TypedDict):
    """A GeoJSON Point geometry."""

    type: Literal["Point"]

    coordinates: tuple[float, float]
    """The coordinates of the point, in [longitude, latitude] order."""


class GeocoderFeature(TypedDict):
    """The expected response from a geocoder query.

    This must be a GeoJSON Point feature, extended with some additional properties.
    """

    type: Literal["Feature"]

    id: NotRequired[Any]

    properties: dict[str, Any]

    geometry: GeoJsonPoint
    """A GeoJSON Point geometry representing the location of the geocoder result."""

    text: str
    """Text representing the feature (e.g. "Austin")."""

    language: NotRequired[str]
    """The language code of the text returned in text."""

    place_name: str
    """Human-readable text representing the full result hierarchy.

    For example, "Austin, Texas, United States".
    """

    place_type: Sequence[str]
    """An array of index types that this feature may be returned as.

    Most features have only one type matching its id.
    """

    bbox: NotRequired[tuple[float, float, float, float]]
    """Bounding box of the form [minx,miny,maxx,maxy]."""

    center: NotRequired[tuple[float, float]]
    """The center of the feature [lng, lat]."""


class GeocoderFeatureCollection(TypedDict):
    """A collection of geocoder features."""

    type: Literal["FeatureCollection"]

    features: Sequence[GeocoderFeature]
    """The geocoder features returned from the query."""


class GeocoderHandler(Protocol):
    """A protocol for handling geocoder queries from the frontend."""

    async def __call__(
        self,
        query: str,
    ) -> GeocoderFeatureCollection | GeocoderFeature | None:
        """Handle a geocoder query.

        Args:
            query: The geocoder query string from the frontend.

        Returns:
            A `FeatureCollection` or `Feature` to return to the frontend. Return a FeatureCollection if there are multiple results found.

        """
        ...


class GeocoderControl(BaseControl):
    """A deck.gl GeocoderControl.

    Passing this to [`Map.controls`][lonboard.Map.controls] will add a geocoder search
    box to the map.
    """

    _pending_tasks: set[asyncio.Task[None]]
    _client: GeocoderHandler

    def __init__(self, *, client: GeocoderHandler, **kwargs: Any) -> None:
        """Create a new GeocoderControl.

        Keyword Args:
            client: An async callable that takes a geocoder query string and returns a `Feature` or `FeatureCollection` representing the geocoder results. Use the GeocoderHandler protocol for type checking.
            kwargs: Other keyword arguments passed to the `BaseControl`.

        """
        self._client = client
        self.on_msg(_on_geocoder_dispatch)
        self._pending_tasks = set()

        super().__init__(**kwargs)

    @classmethod
    def from_geopy(cls, geocoder: Geocoder, **kwargs: Any) -> GeocoderControl:
        """Create a GeocoderControl from a [geopy][geopy] geocoder instance.

        The geocoder must be created in [async mode] using an async adapter (e.g.
        [`AioHTTPAdapter`][geopy.adapters.AioHTTPAdapter]).

        [async mode]: https://geopy.readthedocs.io/en/stable/#async-mode

        **Example:**

        ```py
        from lonboard.controls import GeocoderControl
        from lonboard import Map
        from geopy.adapters import AioHTTPAdapter
        from geopy.geocoders import Nominatim

        geocoder = Nominatim(user_agent="lonboard", adapter_factory=AioHTTPAdapter)
        geocoder_control = GeocoderControl.from_geopy(geocoder)

        m = Map(layer, controls=[geocoder_control])
        ```
        """
        from geopy.adapters import BaseAsyncAdapter
        from geopy.location import Location

        if not isinstance(geocoder.adapter, BaseAsyncAdapter):
            msg = (
                "The geopy geocoder must be created in async mode.\n"
                "Pass `adapter_factory=AioHTTPAdapter` to the geocoder constructor.\n"
                "See https://geopy.readthedocs.io/en/stable/#async-mode for more info.",
            )
            raise TypeError(msg)

        def convert_location(location: Location) -> GeocoderFeature:
            return {
                "type": "Feature",
                "properties": {},
                "geometry": {
                    "type": "Point",
                    "coordinates": (location.longitude, location.latitude),
                },
                "text": location.address,
                "place_name": location.address,
                "place_type": ["geopy-result"],
                "center": (location.longitude, location.latitude),
            }

        async def handler(
            query: str,
        ) -> GeocoderFeatureCollection | GeocoderFeature | None:
            location = await geocoder.geocode(query, exactly_one=False)
            if location is None:
                return None

            if isinstance(location, list):
                locations = [convert_location(loc) for loc in location]
                return {
                    "type": "FeatureCollection",
                    "features": locations,
                }

            if isinstance(location, Location):
                return convert_location(location)

            raise TypeError(f"Unexpected geopy result type: {location}")

        return cls(client=handler, **kwargs)

    _control_type = t.Unicode("geocoder")



# ruff: noqa: SLF001, UP031

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, NoReturn, TypeVar

from traitlets.traitlets import Container, TraitError, TraitType, Undefined
from traitlets.utils.descriptions import class_of, describe

if TYPE_CHECKING:
    from traitlets.traitlets import HasTraits
    from traitlets.utils.sentinel import Sentinel

DEFAULT_INITIAL_VIEW_STATE = {
    "latitude": 10,
    "longitude": 0,
    "zoom": 0.5,
    "bearing": 0,
    "pitch": 0,
}


class FixedErrorTraitType(TraitType):
    """A custom subclass of traitlets.TraitType.

    This is because its `error` method ignores the `info` passed in. See
    https://github.com/developmentseed/lonboard/issues/71 and
    https://github.com/ipython/traitlets/pull/884.
    """

    def error(
        self,
        obj: HasTraits | None,
        value: Any,
        error: Exception | None = None,
        info: str | None = None,
    ) -> NoReturn:
        """Raise a TraitError.

        Parameters
        ----------
        obj : HasTraits or None
            The instance which owns the trait. If not
            object is given, then an object agnostic
            error will be raised.
        value : any
            The value that caused the error.
        error : Exception (default: None)
            An error that was raised by a child trait.
            The arguments of this exception should be
            of the form ``(value, info, *traits)``.
            Where the ``value`` and ``info`` are the
            problem value, and string describing the
            expected value. The ``traits`` are a series
            of :class:`TraitType` instances that are
            "children" of this one (the first being
            the deepest).
        info : str (default: None)
            A description of the expected value. By
            default this is infered from this trait's
            ``info`` method.

        """
        if error is not None:
            # handle nested error
            error.args += (self,)
            if self.name is not None:
                # this is the root trait that must format the final message
                chain = " of ".join(describe("a", t) for t in error.args[2:])
                if obj is not None:
                    error.args = (
                        "The '{}' trait of {} instance contains {} which "
                        "expected {}, not {}.".format(
                            self.name,
                            describe("an", obj),
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
                else:
                    error.args = (
                        "The '{}' trait contains {} which expected {}, not {}.".format(
                            self.name,
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
            raise error
        # this trait caused an error
        if self.name is None:
            # this is not the root trait
            raise TraitError(value, info or self.info(), self)
        # this is the root trait
        if obj is not None:
            e = "The '{}' trait of {} instance expected {}, not {}.".format(
                self.name,
                class_of(obj),
                # CHANGED:
                # Use info if provided
                info or self.info(),
                describe("the", value),
            )
        else:
            e = "The '{}' trait expected {}, not {}.".format(
                self.name,
                # CHANGED:
                # Use info if provided
                info or self.info(),
                describe("the", value),
            )
        raise TraitError(e)


T = TypeVar("T")


# TODO: switch to
# class VariableLengthTuple(traitlets.Container[tuple[T, ...]])
# When we can upgrade to traitlets 5.10 (depends on Colab upgrading)
class VariableLengthTuple(Container):
    """An instance of a Python tuple with variable numbers of elements of the same type."""

    klass = list  # type:ignore[assignment]
    _cast_types: Any = (tuple,)

    def __init__(
        self,
        trait: T | Sentinel,
        default_value: tuple[T, ...] | Sentinel | None = Undefined,
        minlen: int = 0,
        maxlen: int = sys.maxsize,
        **kwargs: Any,
    ) -> None:
        """Create a tuple trait type.

        The default value is created by doing ``list(default_value)``,
        which creates a copy of the ``default_value``.

        ``trait`` can be specified, which restricts the type of elements
        in the container to that TraitType.

        If only one arg is given and it is not a Trait, it is taken as
        ``default_value``:

        ``c = List([1, 2, 3])``

        Parameters
        ----------
        trait : TraitType [ optional ]
            the type for restricting the contents of the Container.
            If unspecified, types are not checked.
        default_value : SequenceType [ optional ]
            The default value for the Trait.  Must be list/tuple/set, and
            will be cast to the container type.
        minlen : Int [ default 0 ]
            The minimum length of the input list
        maxlen : Int [ default sys.maxsize ]
            The maximum length of the input list
        kwargs:
            passed on to traitlets.Container.

        """
        self._maxlen = maxlen
        self._minlen = minlen
        super().__init__(trait=trait, default_value=default_value, **kwargs)
        self.tag(sync=True)

    def length_error(self, obj: Any, value: Any) -> None:
        e = "The '%s' trait of %s instance must be of length %i <= L <= %i" % (
            self.name,
            class_of(obj),
            self._minlen,
            self._maxlen,
        )
        e += f", but a value of {value} was specified."
        raise TraitError(e)

    def validate_elements(self, obj: Any, value: Any) -> Any:
        length = len(value)
        if length < self._minlen or length > self._maxlen:
            self.length_error(obj, value)

        trait = self._trait

        validated = []
        for v in value:
            try:
                v = trait._validate(obj, v)  # noqa: PLW2901
            except TraitError as error:
                self.error(obj, v, error)
            else:
                validated.append(v)

        return tuple(validated)



from __future__ import annotations

from io import StringIO
from typing import IO, TYPE_CHECKING, TextIO, overload

from ipywidgets.embed import dependency_state, embed_minimal_html

if TYPE_CHECKING:
    from pathlib import Path

    from lonboard import Map


# HTML template to override exported map as 100% height
_HTML_TEMPLATE = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>{title}</title>
</head>
<style>
    html {{ height: 100%; }}
    body {{ height: 100%; overflow: hidden;}}
    .widget-subarea {{ height: 100%; }}
    .jupyter-widgets-disconnected {{ height: 100%; }}
</style>
<body>
{snippet}
</body>
</html>
"""


@overload
def map_to_html(
    m: Map,
    *,
    filename: None = None,
    title: str | None = None,
) -> str: ...


@overload
def map_to_html(
    m: Map,
    *,
    filename: str | Path | TextIO | IO[str],
    title: str | None = None,
) -> None: ...


def map_to_html(
    m: Map,
    *,
    filename: str | Path | TextIO | IO[str] | None = None,
    title: str | None = None,
) -> str | None:
    def inner(fp: str | Path | TextIO | IO[str]) -> None:
        original_height = m.height
        try:
            with m.hold_trait_notifications():
                m.height = "100%"
                embed_minimal_html(
                    fp,
                    views=[m],
                    title=title or "Lonboard export",
                    template=_HTML_TEMPLATE,
                    drop_defaults=False,
                    # Necessary to pass the state of _this_ specific map. Otherwise, the
                    # state of all known widgets will be included, ballooning the file size.
                    state=dependency_state((m), drop_defaults=False),
                )
        finally:
            # If the map had a height before the HTML was generated, reset it.
            m.height = original_height

    if filename is None:
        with StringIO() as sio:
            inner(sio)
            return sio.getvalue()

    else:
        inner(filename)
        return None



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

import numpy as np

from lonboard._environment import DEFAULT_HEIGHT
from lonboard._serialization import serialize_view_state
from lonboard.traits._base import FixedErrorTraitType
from lonboard.traits._upstream_wrappers import Unicode
from lonboard.view_state import BaseViewState, MapViewState

if TYPE_CHECKING:
    from traitlets.traitlets import HasTraits

    from lonboard._map import Map


class BasemapUrl(Unicode):
    """Validation for basemap url."""

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)

    def validate(self, obj: HasTraits | None, value: Any) -> Any:
        value = super().validate(obj, value)

        try:
            parsed = urlparse(value)
        except:  # noqa: E722
            self.error(obj, value, info="to be a URL")

        if not parsed.scheme.startswith("http"):
            self.error(obj, value, info="to be a HTTP(s) URL")

        return value


class MapHeightTrait(FixedErrorTraitType):
    """Trait to validate map height input."""

    allow_none = True
    default_value = DEFAULT_HEIGHT

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True)

    def validate(self, obj: Any, value: Any) -> str:
        if isinstance(value, (int, np.integer)):
            return f"{value}px"

        if isinstance(value, str):
            return value

        self.error(obj, value)
        assert False


class ViewStateTrait(FixedErrorTraitType):
    """Trait to validate view state input."""

    allow_none = True
    default_value = None

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, to_json=serialize_view_state)

    def validate(self, obj: Map, value: Any) -> None | BaseViewState:
        if value is None:
            return None

        if isinstance(value, BaseViewState):
            return value

        # Otherwise dict input
        view = obj.view
        validator = view._view_state_type if view is not None else MapViewState  # noqa: SLF001

        # The frontend currently sends back data in camelCase
        snake_case_kwargs = {_camel_to_snake(k): v for k, v in value.items()}
        return validator(**snake_case_kwargs)  # type: ignore


def _camel_to_snake(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()



from __future__ import annotations

from typing import Generic, TypedDict, TypeVar

from traitlets import HasTraits, TraitType

Trait = TypeVar("Trait", bound=TraitType)
Value = TypeVar("Value")
Owner = TypeVar("Owner", bound=HasTraits)


class TraitProposal(TypedDict, Generic[Trait, Value, Owner]):
    """The type of a traitlets proposal.

    The input into a `@validate` method.
    """

    trait: Trait
    value: Value
    owner: Owner



from __future__ import annotations

from typing import TYPE_CHECKING

import lonboard.traits as t
from lonboard.layer._base import BaseArrowLayer

if TYPE_CHECKING:
    import sys

    from arro3.core.types import ArrowStreamExportable

    from lonboard.types.layer import ArcLayerKwargs, PointAccessorInput

    if sys.version_info >= (3, 12):
        from typing import Unpack
    else:
        from typing_extensions import Unpack


class ArcLayer(BaseArrowLayer):
    """Render raised arcs joining pairs of source and target coordinates."""

    _layer_type = t.Unicode("arc")

    table = t.ArrowTableTrait(geometry_required=False)
    """A GeoArrow table.

    This is the fastest way to plot data from an existing GeoArrow source, such as
    [geoarrow-rust](https://geoarrow.github.io/geoarrow-rs/python/latest) or
    [geoarrow-pyarrow](https://geoarrow.github.io/geoarrow-python/main/index.html).
    """

    great_circle = t.Bool(None, allow_none=True)
    """If `True`, create the arc along the shortest path on the earth surface.

    - Type: `bool`, optional
    - Default: `False`
    """

    num_segments = t.Int(None, allow_none=True)
    """The number of segments used to draw each arc.

    - Type: `int`, optional
    - Default: `50`
    """

    width_units = t.Unicode(None, allow_none=True)
    """
    The units of the line width, one of `'meters'`, `'common'`, and `'pixels'`. See
    [unit
    system](https://deck.gl/docs/developer-guide/coordinate-systems#supported-units).

    - Type: `str`, optional
    - Default: `'pixels'`
    """

    width_scale = t.Float(None, allow_none=True, min=0)
    """
    The scaling multiplier for the width of each line.

    - Type: `float`, optional
    - Default: `1`
    """

    width_min_pixels = t.Float(None, allow_none=True, min=0)
    """The minimum line width in pixels.

    - Type: `float`, optional
    - Default: `0`
    """

    width_max_pixels = t.Float(None, allow_none=True, min=0)
    """The maximum line width in pixels.

    - Type: `float`, optional
    - Default: `None`
    """

    get_source_position = t.PointAccessor(None, allow_none=True)
    """Source position of each object
    """

    get_target_position = t.PointAccessor(None, allow_none=True)
    """Target position of each object
    """

    get_source_color = t.ColorAccessor(None, allow_none=True)
    """Source color of each object
    """

    get_target_color = t.ColorAccessor(None, allow_none=True)
    """Target color of each object
    """

    get_width = t.FloatAccessor(None, allow_none=True)
    """The line width of each object, in units specified by `widthUnits`.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `1`.
    """

    get_height = t.FloatAccessor(None, allow_none=True)
    """Height color of each object
    """

    get_tilt = t.FloatAccessor(None, allow_none=True)
    """
    Use to tilt the arc to the side if you have multiple arcs with the same source and
    target positions.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the width for all paths.
        - If an array is provided, each value in the array will be used as the width for
          the path at the same row index.
    - Default: `0`.
    """

    def __init__(
        self,
        table: ArrowStreamExportable,
        *,
        get_source_position: PointAccessorInput,
        get_target_position: PointAccessorInput,
        _rows_per_chunk: int | None = None,
        **kwargs: Unpack[ArcLayerKwargs],
    ) -> None:
        """Construct an ArcLayer from existing Arrow data.

        Args:
            table: An Arrow table from any Arrow-compatible library. This does not need to have a geometry column as geometries are passed separately in `get_source_position` and `get_target_position`.

        Keyword Args:
            get_source_position: Source position of each object.
            get_target_position: Target position of each object.
            kwargs: Extra args passed down as ArcLayer attributes.

        """
        super().__init__(
            table=table,
            _rows_per_chunk=_rows_per_chunk,
            get_source_position=get_source_position,  # type: ignore
            get_target_position=get_target_position,  # type: ignore
            **kwargs,
        )
